#
# $Id: cert.py,v 1.17 2006/05/19 23:50:35 mwalfish Exp $
# utilites for working with certificates 
#
#
# Copyright (C) 2005-2006 Michael Walfish
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import base64
import binascii
from dqm.util.convert import *
import dqm.cryptkey_wrap.cryptkey as cryptkey
import datetime
import dqm.err as dqm_err

### XXX: this interface is weird and sucky. It needs to be fixed by:
### creating a separate base.py file, moving the simplecert stuff into 
### its own file, and moving the create/verify functions outside of the
### object. basically, our current encapsulation scheme is lame.

class BaseCert:

    """This class is abstract because we assume that derived
    classes could have their own formats for the canonical layout of the 
    cert. The main purpose of this class is to handle the details of
    marshalling between our software's internal representation of a
    certificate and a string format suitable for network transmission or
    disk storage."""
    def create(self, signer_id, pub_key, quota, expiry, priv_key):
	"""signer_id is the id of the authority to which is associated
	      the private key, priv_key
	   pub_key is the public key to be signed.
	   quota is the quota being allocated by the certificate signer
	   expiry is the time at which the certificate is no longer valid
	   priv_key is the private key of the certificate signer.
	returns a string that is a certificate that can be passed to
	verify_cert or any other certification verifier that knows about
	the format the particular derived class used"""
	raise AssertionError, 'Abstract method called'

    def verify(self, st, pub_key):
	"""st is a stringized certificate 
	returns an error code. 0 for success. other values indicate
	failure.
	"""
	raise AssertionError, 'Abstract method called'

    def str_to_cert(self, st):
	"""Takes as input a string representing the whole certificate (the
	same type of input to the verify() method) and returns a dictionary 
	containing the fields of the certificate. Callers should use this if 
	they want to query on any of the fields in the certificate. Those 
	fields are, for now:
	    'signer_id', 'pub_key', 'quota', 'expiry', 'sig',
	    'info_blob' (the latter being the thing that was signed)"""
	raise AssertionError, 'Abstract method called'

class SimpleCert(BaseCert):
    """Simple certificate format. In the future we may want a version that
    stores certificates in the standard formats like PKCS #1, for
    example. Hopefully, the BaseCert interface won't have to change."""

    """The format is: 
	 info_blob:sig
     sig is a base64 representation of the signature of the info_blob
     info_blob is:
	 signer_id:pub_key:quota:expiry 
     signer_id is a string that may not contain the ':' character
     pub_key is written in base64
     quota is an integer, in ASCII string representation, e.g., "135"
     expiry is a date written as YYYY-MM-DD"""
    
    DELIM = ':'

    def create(self, signer_id, pub_key, quota, expiry, priv_key):
	""" arg types:
	     signer_id: a python string not containing a ':'
	     pub_key: a Key
	     quota:   an integer
	     expiry:  a Python date object
	   will raise ValueError if arguments are not the right type"""

	DELIM = SimpleCert.DELIM
	if signer_id.find(DELIM) != -1: 
	    raise ValueError, "%s in signer_id" % DELIM
	if type(quota) is not int:
	    raise ValueError, "quota has to be an integer"

	pub_key_str = pub_key.to_str()
	quota_str   = str(quota)
	expiry_str  = expiry.isoformat()
	
	info_blob = DELIM.join([signer_id, 
			       pub_key_str, 
			       quota_str,
	                       expiry_str])

	sig = priv_key.sign(info_blob)
    
	sig_64 = base64.encodestring(sig)
	sig_64 = sig_64.replace('\n','')
	cert_blob = DELIM.join([info_blob, sig_64])
	
	return cert_blob

    def verify(self, st, pub_key):
    
	try: cert = self.str_to_cert(st)	
	except (ValueError, binascii.Error): return dqm_err.WRONG_FORMAT

	date_valid = (datetime.datetime.utcnow() < cert['expiry'])
	if date_valid is False: return dqm_err.EXPIRED_CERT

	sig_valid   = pub_key.verify(cert['info_blob'], cert['sig'])
	if sig_valid == 1: return dqm_err.SUCCESS
	else: return dqm_err.INVALID_CERT_SIG

    # to query the individual fields of a cert. Can throw ValueError
    # if the cert is in the wrong format
    def str_to_cert(self, st):
	DELIM = SimpleCert.DELIM

	sig_fields  = st.split(DELIM)
	info_blob   = DELIM.join(sig_fields[:-1])
	
	signer_id, pub_key_str, quota_str, expiry_str, sig_64 = sig_fields
	year,month,day = expiry_str.split('-')
	expiry = datetime.date(int(year),int(month),int(day))
	quota  = int(quota_str)
	pub_key = cryptkey.str_to_pubkey(pub_key_str)
	sig     = base64.decodestring(sig_64)

	cert = {'signer_id': signer_id, 
	        'pub_key'  : pub_key,
	        'quota'    : quota,
	        'expiry'   : expiry,
		'sig'      : sig,
		'info_blob' : info_blob}
	
	return cert
