# 
# $Id: disposable.py,v 1.16 2006/05/19 23:50:35 mwalfish Exp $
#  utilities for working with disposables. The disposable is the piece of the 
#  stamp that is 'single-use'

#
# Copyright (C) 2005-2006 Michael Walfish
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import base64
import binascii
from dqm.util.convert import *
import datetime
import dqm.err as dqm_err

class BaseDisposable:

    """This class is similar to the BaseCertMgr -- it mainly handles the
    details of marshalling between our software's internal representation
    of a disposable and a string format suitable for network transmission.
    It is designed to be 'pluggable' in the sense that another format can
    be substituted by just making a different derived class."""

    def create(self, i, priv_key, weekstamp=None):
	""" create a stringized version of a 'disposable'
	'i' is the counter, and priv_key is the private key that will sign
	the disposable. The default is to create a disposable with the
	current week stamped, but one can ask this method for other weeks
	as well. Note that it is the responsibility of the caller to
	actually manage the 'i' values -- i.e., making sure of things like
	that the same disposable isn't used twice."""
	raise AssertionError, 'Abstract method called'

    def verify(self, st, pub_key):
	"""st is a stringized disposable. this function returns true if the
	disposable is valid in the sense of its signature being valid and
	its weekstamp being valid."""
	raise AssertionError, 'Abstract method called'

    def str_to_disposable(self, st):
	"""Takes as input a string representing a disposable (the same type
	of input to the verify() method) and returns a dictionary
	containing the fields of the disposable. Callers should use this if
	they want to query on any of the fields in the disposable. Those
	fields are, for now:
	    'i', 'weekstamp', 'sig', 'info_blob'
	the info_blob is the string that was signed to produce the
	signature"""
	raise AssertionError, 'Abstract method called'

class SimpleDisposable:

    """Simple format for a disposable. Follows in the ':' and base-64
    ethos of the SimpleCert in cert.py"""

    """The format is:
	 info_blob:sig
    sig is a base64 rperesentation of the signature of the info_blob
    info_blob is:
	i:weekstamp
    i is an integer
    weekstamp is an integer representing the current week (using the iso
    calendar in which there are 52 weeks in the year)"""

    DELIM = ':'

    def create(self, i, priv_key, weekstamp=None):
	"""arg types:
	    i: an integer
	    priv_key: a PrivKey
	    weekstamp: an integer
	  will raise ValueError if arguments are not the right types"""
	DELIM = SimpleDisposable.DELIM 

	if (type(i) is not int) and (type(i) is not long):
	    raise ValueError, "i has to be an integer"

	if weekstamp is None:
	    weekstamp = datetime.datetime.utcnow().isocalendar()[1]
	elif type(weekstamp) is not int:
	    raise ValueError, "weekstamp has to be an integer"

	info_blob = DELIM.join([str(i), str(weekstamp)])
	sig       = priv_key.sign(info_blob)
	sig_64    = base64.encodestring(sig)
	sig_64    = sig_64.replace('\n','')
	disposable_str = DELIM.join([info_blob, sig_64])

	return disposable_str

    def verify(self, st, pub_key):

	try: disposable = self.str_to_disposable(st)
	except (ValueError, binascii.Error): return dqm_err.WRONG_FORMAT

	## To validate the weekstamp, we have to see whether the 
	## weekstamp is from this week or the previous. the only wrinkle 
	## is that if we're currently in week 1, we don't know whether the
	## previous week was 52 or 53, since ISO year varies. so special
	## case the handling of week 1.
	today         = datetime.datetime.utcnow()
	curr_iso_week = today.isocalendar()[1]

	if curr_iso_week == 1:
	    one_week_ago = today - datetime.timedelta(days=7)
	    prev_iso_week = one_week_ago.isocalendar()[1]
	else:
	    prev_iso_week = curr_iso_week - 1

	wkstamp    = disposable['weekstamp']
	date_valid = (wkstamp is curr_iso_week) or \
		     (wkstamp is prev_iso_week)
	if date_valid is False: return dqm_err.BAD_WEEKSTAMP

	sig_valid = pub_key.verify(disposable['info_blob'], disposable['sig'])
	if sig_valid == 1: return dqm_err.SUCCESS
	else: return dqm_err.INVALID_DISPOSABLE_SIG

    def str_to_disposable(self, st):

	DELIM = SimpleDisposable.DELIM

	disposable_fields = st.split(DELIM)
	info_blob         = DELIM.join(disposable_fields[:-1])

	i_str, weekstamp_str, sig_64 = disposable_fields
	i, weekstamp = int(i_str), int(weekstamp_str)
	sig = base64.decodestring(sig_64)

	disposable = {'i'         : i,
		      'weekstamp' : weekstamp,
		      'sig'       : sig,
		      'info_blob' : info_blob}
		    
	return disposable
