/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#ifndef _CRYPTKEY_H_
#define _CRYPTKEY_H_

#include "crypt.h"
#include <openssl/rsa.h>

class str;
class pubkey;
class privkey;
class rabin_pubkey;
class rabin_privkey;
class rsa_pubkey;
class rsa_privkey;

// import functions. 
//
// right now, these aren't fancy. in particular, the private key file isn't
// password-protected and salted, but it wouldn't be hard to get
// these functions, given what DM and Max have already.
//
// also, we are not supporting every concievable import and export method.
// rather, we are just supporting the ones that the current DQM code needs:
//   --moving private keys to and from files
//   --moving public keys to and from files
//   --moving public keys to and from base64-encoded-strings
//
ptr<pubkey> str_to_pubkey(const str& s);
ptr<pubkey> file_to_pubkey(const str& fname);
ptr<privkey> file_to_privkey(const str& fname);

ptr<rabin_privkey> gen_rabin_privkey(int nbits);
ptr<rsa_privkey> gen_rsa_privkey(int nbits, unsigned long e = 3);

class pubkey {
   
    public:
	virtual bool verify(const str& msg, const str& sig) const = 0;
	virtual str to_str() const = 0;
	virtual bool to_file(const str& fname) const {
	    str s = to_str();
	    return str2file(fname, s, 0644, true);
	}
	virtual ~pubkey() { }
};

class privkey {

    public:
	virtual str to_str() const = 0;
	virtual bool to_file(const str& fname) const {
	    str s = to_str();
	    return str2file(fname, s, 0600, true);
	}
	virtual str sign(const str& msg) const = 0;

	virtual ptr<pubkey> to_pubkey() const = 0;

	virtual ~privkey() {}
};

// Rabin

class rabin_pubkey : public pubkey {

    public:
        static ptr<rabin_pubkey> from_str(const char* p, int len);

	rabin_pubkey(ptr<rabin_pub> pubk) : pubk(pubk) {}
	virtual ~rabin_pubkey() { }

	str  to_str() const;
	bool verify(const str& msg, const str& sig) const;

    private:
	// no copying
	rabin_pubkey(const rabin_pubkey&);
	const rabin_pubkey& operator= (const rabin_pubkey&);

	ptr<rabin_pub> pubk;
};

class rabin_privkey : public privkey {

    public:
	static ptr<rabin_privkey> from_str(const str& s);

	rabin_privkey(ptr<rabin_priv> privk) : privk(privk) {}
	virtual ~rabin_privkey() { }

	str sign(const str& msg) const;
	ptr<pubkey> to_pubkey() const {
	    // the privk is a rabin_pub object
	    return New refcounted<rabin_pubkey> (privk);
	}

	str to_str() const;

    private:
	// no copying
	rabin_privkey(const rabin_privkey&);
	const rabin_privkey& operator= (const rabin_privkey&);

	ptr<rabin_priv> privk;
};

// RSA
bool get_digest(mstr& digest, int siglen, const str& msg, RSA* rsa);

// the RSA pub and priv objects are responsible for freeing the RSA* that is 
// passed into them.
class rsa_pubkey : public pubkey {

    public:

	static ptr<rsa_pubkey> from_str(const char* p, int len);

	rsa_pubkey(RSA* rsa) : rsa(rsa) {}
	virtual ~rsa_pubkey() { RSA_free(rsa); }

	str to_str() const;
	bool verify(const str& msg, const str& sig) const;

    private:
	// no copying
	rsa_pubkey(const rsa_pubkey&);
	const rsa_pubkey& operator= (const rsa_pubkey&);

	RSA* rsa;
};

// the rest of the RSA functions are unimplemented for now b/c
// we're not doing these functions in C right now.
class rsa_privkey : public privkey {

    public:

	rsa_privkey(RSA* rsa) : rsa(rsa) {}
	virtual ~rsa_privkey() { RSA_free(rsa); }

	str sign(const str& msg) const;

	ptr<pubkey> to_pubkey() const {

	    // copy the RSA object and hand it to a new
	    // rsa_pubkey object
	    RSA* rsa_new = RSA_new();
	    BIGNUM* n = BN_dup(rsa->n);
	    BIGNUM* e = BN_dup(rsa->e);

	    if (!rsa_new || !n || !e) {
		// these 'free' functions can accept NULL argument.
		RSA_free(rsa_new);
		BN_free(n);
		BN_free(e);
		return NULL;
	    }

	    rsa_new->n = n;
	    rsa_new->e = e;

	    return New refcounted<rsa_pubkey>(rsa_new);	
	}

	// In this implementation, we don't have a canonical stringized
	// version of an RSA private key. RSA private keys are simply read
	// from and written to files directly.
	str to_str() const { assert(false); }

	// overrides the default implementation given in the base
	// class. we do this overriding because RSA private keys are not 
	// stored on disk as simply stringized versions of themselves
	bool to_file(const str& fname) const;
	    
    private:
	// no copying
	rsa_privkey(const rsa_privkey&);
	const rsa_privkey& operator= (const rsa_privkey&);
	
	RSA* rsa;
};

#endif
