#
# $Id: dqm_client.py,v 1.8 2006/05/19 23:50:35 mwalfish Exp $
#
#   Interface to a DQM portal. Takes care of the RPCs and contacting
#   several portals.
#
#
#
# Copyright (C) 2005-2006 Michael Walfish
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import struct
import dqm.rpc.rpc_udp as rpc_udp
import dqm.svc.dqm_portal as dqm_portal
import dqm.user.process_res as process_res

pack = struct.pack

codes = { "DQM_OK" : 0,
          "DQM_NOT_FOUND" : 1,
          "DQM_BAD_METASTAMP" : 2,
          "DQM_METASTAMP_REUSED" : 4,
          "DQM_SEEN_KEY" : 8,
          "DQM_NO_MORE_STORAGE" : 16,
          "DQM_BAD_PARAM" : 32 }
strcodes = {}

for k in codes.keys():
    strcodes[codes[k]] = k

# pack the key, val, and meta-stamp into the RPC format expected by the
# server.
def pack_set_args(key, val, ms):

    # argument for a 'set' is: 
    #  key 
    #  val 
    #  ms_len 
    #  ms
    ms_len = len(ms)

    # 4-byte alignment
    leftover = ms_len % 4
    if leftover:   padding = 4 - leftover
    else:          padding = 0
	
    fmt_string = '>20s20si%ds%dx' % (ms_len, padding)
    return pack(fmt_string, key, val, ms_len, ms)


class DQM:

    def __init__(self, portals):
	"""portals should be a list of (IP,port) tuples. IP can be
	a name or a dotted 4 string, and port should be numeric."""
	self.rpc_cli = rpc_udp.rpc_client()
	self.portals = portals

    def test(self, key, val, cb):
	"""Does a DQM TEST request to check if 'key' is stored and if the
	value equals val. The function 'cb' will be called back with a tuple. 
	The first entry is True or False representing the success of the test. 
	If the second entry is not the empty string, then it is a string
	describing the error that occurred."""
   
	which_portal = 0
	self.make_test_req(which_portal, key, val, cb)

    def set(self, key, val, ms, cb):
	"""Does a DQM SET request to put 'val' in the DQM under 'key', 
	by using metastamp ms. When the request is complete, 'cb' will be 
	called with a (boolean,error) tuple. The boolean is 'True' if the
	SET request was successful and False otherwise. The second entry in
	the tuple will indicate what went wrong."""
	which_portal = 0
	self.make_set_req(which_portal, key, val, ms, cb)

    def make_test_req(self, which, key, val, cb):
	arg = key
	self.rpc_cli.call(dqm_portal,
			  dqm_portal.DQM_PORTAL_PROGRAM,
			  dqm_portal.DQM_PORTAL_VERSION,
			  dqm_portal.DQM_PORTAL_PROC_TEST,
			  arg, 
			  self.portals[which], # (host,port)
			  10.0, # 10 sec. timeout
			  self.test_cb,
			  (which,key,val,cb))

    def test_cb(self, (which_portal, key, val, cb), (rc, res_h) ):

	if rc == rpc_udp.CLI_CALL_COMPLETED:
	    res = process_res.process_res(res_h)
	    if res.status == dqm_portal.DQM_OK:
		if res.val == val:
		    cb((True, ''))
		    return
	    else:
		cb((False, ''))
		return

	# Note that we fall through to the code below if the portal
	# returned a bogus value and claimed it was right. If that happens
	# the caller really should remove the portal because it means the
	# portal isn't doing its job.

	self.try_next(which_portal, cb,
	      lambda portal: self.make_test_req(portal, key, val, cb))


    def make_set_req(self, which, key, val, ms, cb):

	arg = pack_set_args(key, val, ms) 

	self.rpc_cli.call(dqm_portal,
			  dqm_portal.DQM_PORTAL_PROGRAM,
			  dqm_portal.DQM_PORTAL_VERSION,
			  dqm_portal.DQM_PORTAL_PROC_SET,
			  arg, 
			  self.portals[which], # (host,port)
			  15.0, # 15 sec. timeout
			  self.set_cb,
			  (which,key,val,ms,cb))

    def set_cb(self, (which_portal, key, val, ms, cb), (rc, res) ):

	if rc == rpc_udp.CLI_CALL_COMPLETED:
	    if res == dqm_portal.DQM_OK: cb((True, ''))
	    else:                        cb((False,strcodes[res]))
	    return

	self.try_next(which_portal, cb,
	      lambda portal: self.make_set_req(portal, key, val, ms, cb))


    # This function exists to avoid repeating code. It makes the functions
    # harder to read, but in this case the trade-off is worth it.
    def try_next(self, which_portal, cb, next_op):

	which_portal = which_portal + 1
	if which_portal < len(self.portals): next_op(which_portal)
	else: cb((False, 'no portal could be contacted'))
