/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#ifndef _DQMNODE_H_
#define _DQMNODE_H_

#include "async.h"
#include "dqm_types.h"
#include "qhash.h"

#define HASH_SIZE      20
#define LINUX_ALIGN   512

struct storer_info {
    str hostname;
    u_int32_t ip_addr;
    u_int16_t storer_port;
    u_int16_t internal_port;
    sockaddr_in sockaddr;
};

struct ip_port {
    u_int32_t ip;
    u_int16_t port;
};

inline bool
operator== (const ip_port &a, const ip_port &b)
{
    return a.port == b.port && a.ip == b.ip;
}

inline bool
operator!= (const ip_port &a, const ip_port &b)
{
    return a.port != b.port || a.ip != b.ip;
}

inline const strbuf &
operator<< (const strbuf& sb, const ip_port& a)
{
    struct in_addr addr;
    addr.s_addr = a.ip;
    return sb << inet_ntoa(addr) << ":" << ntohs(a.port);
}

inline const strbuf &
operator<< (const strbuf& sb, const sockaddr_in& a)
{
    return sb << inet_ntoa(a.sin_addr) << ":" << ntohs(a.sin_port);
}

// For the index_map, I am not using DM's hash_bytes in libasync, which
// was an algorithm from Daniel J. Bernstein; that hash function did not
// give good randomness on dqm_key objects (which are themselves random
// 20-byte strings). Instead, I am using a hash function I found at:
//    http://www.cs.yorku.ca/~oz/hash.html 
// That URL claims: 
//    "this algorithm was created for sdbm (a public-domain
//    reimplementation of ndbm) database library. it was found to do
//    well in scrambling bits, causing better distribution of the keys
//    and fewer splits. it also happens to be a good general hashing
//    function with good distribution. the actual function is hash(i) =
//    hash(i - 1) * 65599 + str[i]; what is included below is the faster
//    version used in gawk. [there is even a faster, duff-device
//    version] the magic constant 65599 was picked out of thin air while
//    experimenting with different constants, and turns out to be a
//    prime. this is one of the algorithms used in berkeley db (see
//    sleepycat) and elsewhere."
inline u_int32_t
sdbm_hash_bytes(const char* p, size_t len, u_int32_t seed = 0)
{
    u_int32_t hash = seed;
    const char* end = p + len;

    for (const char* key = p; key < end; key++) 
	hash = *key + (hash << 6) + (hash << 16) - hash;

    return hash;
}

template<> 
struct hashfn<ip_port> {
    hash_t operator() (const ip_port &a) const
   { return ntohs (a.port) << 16 ^ htonl (a.ip); }
};

template<>
struct hashfn<ptr<dqm_key> > {
    hashfn() {}
    hash_t operator() (const ptr<dqm_key>& a) const {
	return hash_bytes(a->base(), a->size());
    }
};

template<>
struct equals<ptr<dqm_key> > {
    equals() {}
    bool operator() (const ptr<dqm_key>& a, const ptr<dqm_key>& b) const
    {
	if (a->size() != b->size())
	    return false;
	return !memcmp(a->base(), b->base(), a->size());
    }
};

typedef qhash<str, storer_info>  approved_list;

typedef callback<void, ptr<dqm_val> >::ref getcb;

#endif
