/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#ifndef _INDEX_MAP_H_
#define _INDEX_MAP_H_

#include "async.h"
#include "qhash.h"
#include "dqm_node.h"

// We wrap the underlying value up in an object to prevent stupid
// overflow errors. We're primarily trying to avoid having the index_t
// be multiplied with another 32 bit number without an explicit cast to an
// off_t. 
struct index_t {

    index_t() : val(0)     {}
    index_t (u_int32_t v) : val(v)     {}

    // Can assign a 32-bit quantity to an index_t.
    index_t& operator=(u_int32_t v)   { val = v;  return *this; }

    // Note that since the compiler's default assignment operator is in
    // effect, the above two declarations are slightly redundant. We
    // don't need the operator=. The reason is that if we had code like:
    //          index_t a; 
    //          a = 0;
    // then the second constructor will create an index_t from a 32-bit
    // int (0, in this case), and the compiler will then use the assignment 
    // operator for index_t objects.

    // Copying allowed. Use the compiler's defaults for the copy
    // constructor and the assignment operator.
    // index_t(const index_t &a);
    // const index_t& operator=(const index_t& a);

    // cast operators.
    operator off_t() const       { return static_cast<off_t>(val); }
    operator u_int32_t() const   { return val;                     }

    // prefix and postfix increments, respectively
    index_t& operator++()   { ++val; return *this; }
    index_t operator++(int) { index_t temp(*this); ++val; return temp; }
    
    index_t& operator+=(int i) { val += i; return *this; }

    bool operator==(const index_t& a) const { return val == a.val; }
    bool operator< (const index_t& a) const { return val < a.val;  }

    u_int32_t val;
};

// This object maps a dqm_key to an index. Its goal is to store many
// indices while being space efficient. Because two keys can have the 
// 8-bit "checksums" and occupy the same 'slot', the underlying array
// gives false positives sometimes. To address these, we store the
// colliding keys in an overflow table. The number of keys so stored is
// expected to be small.
class index_map {

	// don't tell the world that we're storing indices concatenated
	// with an 8 bit "checksum".
	typedef u_int32_t entry_t;

    public:

	static const index_t EMPTY_ENTRY;

	index_map(u_int32_t num_entries) :
	   tab(New entry_t[num_entries]),
	   num_buckets(num_entries), 
	   hashseed1(rand()),
	   hashseed2(rand()),
	   overflows(0) {
	      memset(tab, static_cast<u_int32_t>(EMPTY_ENTRY), 
	             num_buckets*sizeof(entry_t));
	    }
	~index_map() { 
	    delete [] tab; 
	}


	index_t lookup(ptr<dqm_key>);

	// Note: insert() is *not* idempotent. If you call it three
	// times, you'll wind up with the key in the overflow table
	// mapped to the most recent index. This costs memory, but not a lot,
	// relative to how many keys we are going to store.
	void insert(ptr<dqm_key>, index_t index);

	u_int32_t getoverflows() { return overflows; }

    private:
    
	typedef qhash<ptr<dqm_key>, index_t> overflow_map;

	u_int8_t EXTRACT_CKSUM(entry_t entry);
	index_t  EXTRACT_INDEX(entry_t entry);
	entry_t  COMBINE(u_int8_t ck, index_t index);

	u_int8_t  cksum(ptr<dqm_key>);
	u_int32_t get_loc(ptr<dqm_key> k);

	overflow_map overflow;
	entry_t* tab;
	u_int32_t num_buckets;
	u_int32_t hashseed1;  // for unpredictability
	u_int32_t hashseed2;  // likewise
	u_int32_t overflows;
};


#endif
