/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#include <math.h>
#include "sha1.h"
#include "nodetable_notree.h"
#include "dqm_types.h"
#include "util.h"
#include "log.h"

nodetable_notree::nodetable_notree(int r, u_int32_t num_aliases) :
    naliases(num_aliases), p(p), r(r), nnodes(0)
{   
    memset(&local_id, 0, sizeof(local_id));
} 

#if 0
void 
nodetable_notree::recalc_k()
{
    double k_unrounded = log(nnodes) / log(1. / p); // log_{1/p} m
    // round to integer above current value
    r = (int)(k_unrounded + 1.0);
    if (nnodes > 0 && r == 0) r = 1;
    if (nnodes == 0) r = 0;

    if (r <= 1) return;
}
#endif

nodetable_notree::~nodetable_notree()
{
    ringmember_lstelm* curr;

    for (curr = lst.first(); curr; curr = lst.first()) {
	lst.remove(curr->alias);
	delete curr;
    }
}

void
nodetable_notree::set_local_node(const node_id& id)
{
    local_id = id;
}

void
nodetable_notree::output_params(ptr<base_logger> logger)
{
    char buf[256];
    sprintf(buf, "r: %d. number of nodes: %d.", r, nnodes);
    str s = buf;
    logger->log(s.cstr());
}

ptr<node_list>
nodetable_notree::r_nodes(ptr<dqm_key> key)
{
    if (nnodes < r) 
	fatal("num_nodes = %d. r = %d. r set too high.\n", nnodes, r);

    node_id succ_id;
    node_alias key_as_alias 
       = hash2alias(reinterpret_cast<u_char*>(key->base()));
    ringmember_lstelm* succ = get_succ(key_as_alias, &lst);

    ptr<node_list> ret = New refcounted<node_list>;

    ret->push_back(succ->id);

    // i tracks which hash function we are trying.
    // j tracks the number of unique nodes we have found.
    //   j starts at 1 b/c we already added 1 to the set of elements
    //   we're returning
    for (int i = 0, j = 1; j < r; i++) {

	key_as_alias = h_i_as_alias(key->base(), key->size(), i);

	succ = get_succ(key_as_alias, &lst);
	succ_id = succ->id;

	if (!in_vec(succ_id, ret)) {

	    ret->push_back(succ_id);
	    j++;
	}
    }

    return ret;
}

void
nodetable_notree::add_node(const node_id& id, const str& alias_seed)
{
    alias_vec aliases = get_node_aliases(alias_seed, naliases);
    node_alias alias;
    static int dup_cnt = 0;

    for (u_int32_t i = 0; i < naliases; i++) {

	alias = aliases[i];
	ringmember_lstelm* to_insert = New ringmember_lstelm;

	to_insert->alias = alias;
	to_insert->id    = id;
	if (!lst.insert(to_insert)) {
	    ++dup_cnt;
	    warnx("duplicate id. %d duplicates so far.\n", dup_cnt);
	}
    }

    node_aliases.insert(id, aliases);
    nodes.insert(id, DOWN);
    ++nnodes;
}

// similar to nodetable_geom -- should we try to factor out common
// functions?
void
nodetable_notree::remove_node(const node_id& id)
{
    alias_vec* aliases = node_aliases[id];
    assert(aliases);

    node_alias alias;
    for (u_int32_t i = 0; i < naliases; i++) {

	alias = (*aliases)[i];

	ringmember_lstelm* curr = lst.remove(alias);
	delete curr;
    }
    nodes.remove(id);
    node_aliases.remove(id);
    --nnodes;
}

void
nodetable_notree::dump()
{
    dump_nodes_and_aliases(&nodes, &node_aliases);
    warnx("\nthe ring: ");

    for (ringmember_lstelm* curr = lst.first(); curr; curr = lst.next(curr))
	warnx("%.6x ", curr->alias);

    warnx("\n");
}

// Pick one of the candidate nodes at random. Move to uniform choice,
// whereas previously the choice was weighted.
ptr<node_list>
nodetable_notree::calc_path_put(ptr<node_list> l)
{
    ptr<node_list> ret = New refcounted<node_list>();
    int len = l->size();

    assert(len == r);
    int i = get_rand_int(0, len-1);

    ret->push_back( (*l)[i] );
    return ret;
}


// returns true if id is already in 'vec'
bool
nodetable_notree::in_vec(node_id id, ptr<node_list> vec)
{
    u_int32_t s = vec->size();

    for (u_int32_t i = 0; i < s; i++) 
	if (id == (*vec)[i])
	    return true;

    return false;
}


// hashes the input with the ith hash function, and returns this quantity
// as a 32-bit alias. h_i(x) is just hash(x.i), where:
//                  . means concatenation
//                  i is the ASCII representation of the number i
nodetable::node_alias
nodetable_notree::h_i_as_alias(const char* p, u_int32_t len, int i)
{
    u_char hashed[sha1::hashsize];
    str s(p, len);

    strbuf pre_hash_sb;
    pre_hash_sb << s << i;
    str pre_hash_str(pre_hash_sb);
	  
    sha1_hash(hashed, pre_hash_str.cstr(), pre_hash_str.len());
    return hash2alias(hashed);
}

nodetable::node_alias 
nodetable_notree::hash2alias(const u_char* hash) {
    return ((hash[0] << 24) & 0xff000000) |
           ((hash[1] << 16) & 0x00ff0000) | 
	   ((hash[2] << 8)  & 0x0000ff00) | 
	    (hash[3] & 0x000000ff);
}

