/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#include "aiod.h"
#include "ondisk_storage.h"
#include "dqm_types.h"
#include "util.h"
#include "log.h"
#include "tt.h"
#include "dqm_node.h"
#include "index_map.h"

//#define PRINT_SETTLE_TIMES

#define PATH_DELIM '/'

// This magic number needs to be the first four bytes of the daily
// header. There are two daily headers in a key-value store, one for
// each day (because we store the keys for today and yesterday).
const u_int32_t DQM_MAGIC_NUM = 0x12484812;

// Constants
const int naiod_procs             = 8;
const ssize_t PAIR_SIZE	          = 2 * HASH_SIZE;
const ssize_t FLUSH_SIZE_DEF      = (1 << 17);	
const ssize_t BLOCK_SIZE_DEF      = (1 << 12);    

const u_int32_t HALF_SECOND_us    =  500000;
const u_int32_t TWO_SECONDS_us    = 2000000;
const u_int32_t READ_INTERVAL_us  = TWO_SECONDS_us;
const u_int32_t WRITE_INTERVAL_us = HALF_SECOND_us;
const time_t    READ_INTERVAL_s   = READ_INTERVAL_us / 1000000;
const u_int32_t READ_INTERVAL_ns  = (READ_INTERVAL_us % 1000000)*1000;
const time_t    WRITE_INTERVAL_s  = WRITE_INTERVAL_us / 1000000;
const u_int32_t WRITE_INTERVAL_ns = (WRITE_INTERVAL_us % 1000000)*1000;

const u_int32_t MAX_WRITES_OUT    = 1; 
const u_int32_t MAX_READS_OUT     = naiod_procs + 1;  
// just needs to be a big number
const u_int32_t MAX_QUEUED_WRITES = 0x100000; 
// We can only do, in a very optimistic case, 250 reads per second.
// If the read is more than 10 seconds old, the caller will probably
// time out. So we'll double that number and queue 5000 reads.
//const u_int32_t MAX_QUEUED_READS  = 250000; 
const u_int32_t MAX_QUEUED_READS  = 5000;
const size_t SWITCH_INTERVAL      = 86400;
const ssize_t AIOD_SHM_SIZE       = (1 << 21);  

// XXX: this needs to be configurable instead of hard-coded
const index_t MAX_PAIRS_DAILY  = 0x08000000;     // 128M pairs

// depend on config parameters
ssize_t BLOCK_SIZE;
ssize_t FLUSH_SIZE;
u_int32_t BLOCKS_PER_FLUSH;
u_int32_t PAIRS_PER_BLOCK;
u_int32_t PAIRS_PER_FLUSH;
index_t MAX_FLUSHES_DAILY;
index_t MAX_BLOCKS_DAILY;
ssize_t AIOD_SHM_MAXBUF;
ssize_t HDR_SPACE;

// Prefaces a sequence of key-value pairs. Used to decide
// whether the key-value pairs are fresh and how many of them
// there are.
struct daily_header {
    u_int32_t dqm_magic_num;
    index_t num_blocks;
    index_t max_blocks; // Tells us where the next day's stash is
    epoch_t epoch;
};

struct keyval {
    char key[HASH_SIZE];
    char val[HASH_SIZE];
};

struct kv_block {
    keyval pairs[];
};


// The layout on the disk is:
// [daily_header | padding to HDR_SPACE bytes]
//  [<KV_1><KV_2><KV_{pairs_per_block}> | padding to BLOCK_SIZE]
//  [<KV_>............................> | padding to BLOCK_SIZE]
//   ...
// [daily_header_2 | padding to HDR_SPACE bytes]


#define FILE_FLAGS (O_RDWR | O_DIRECT)
#define KEYVAL_HDR_VALID(hdr)  (hdr.dqm_magic_num == DQM_MAGIC_NUM)

// fwd decls.
inline char* align_up(char* p, ssize_t align);
inline off_t second_half_begins(index_t max_blocks);


bool
power_of_two(u_int32_t num)
{
    int pos;
    int bits = sizeof(u_int32_t)*8;
    u_int32_t num_test = num;
    u_int32_t mask;

    for (pos = 0; pos < bits; pos++) {
	if (num_test & 1)
	    break;
	num_test = num_test >> 1;
    }

    if (pos == bits) {
	assert(num == 0);
	return false;
    }

    mask = 1 << pos;
    return !(~mask & num);
}

void init_disk_constants(int32_t read_size, int32_t write_size)
{
    BLOCK_SIZE = (read_size == -1)  ? BLOCK_SIZE_DEF : read_size;
    FLUSH_SIZE = (write_size == -1) ? FLUSH_SIZE_DEF : write_size; 

    if (!power_of_two(BLOCK_SIZE) || !power_of_two(FLUSH_SIZE))
	fatal("read_size and write_size must be powers of two. "
	      "read_size: %d. write_size: %d\n", BLOCK_SIZE, FLUSH_SIZE);


    // Make sure we don't try to read less than the alignment required
    // by Linux.
    BLOCK_SIZE = max(BLOCK_SIZE, LINUX_ALIGN);
    FLUSH_SIZE = max(FLUSH_SIZE, LINUX_ALIGN);

    BLOCKS_PER_FLUSH  = FLUSH_SIZE / BLOCK_SIZE;
    PAIRS_PER_BLOCK   = BLOCK_SIZE / PAIR_SIZE;
    PAIRS_PER_FLUSH   = PAIRS_PER_BLOCK * BLOCKS_PER_FLUSH;

    MAX_FLUSHES_DAILY   = static_cast<u_int32_t>(MAX_PAIRS_DAILY) / 
					         PAIRS_PER_FLUSH;
    MAX_BLOCKS_DAILY    = static_cast<u_int32_t>(MAX_FLUSHES_DAILY) * 
					         BLOCKS_PER_FLUSH; 
    // need to be able to get a shared memory buffer that accommodates
    // our flush size
    AIOD_SHM_MAXBUF     = max(FLUSH_SIZE_DEF, FLUSH_SIZE); 

    // amount of room to leave before writing key-value pairs to the file.
    // we write the header to a FLUSH_SIZEd piece because writing a
    // FLUSH_SIZEd piece is much faster than writing a smaller amount.
    HDR_SPACE = FLUSH_SIZE;
}

void 
clear_storage(char* pathname)
{
    int fd;
    if ((fd = open(pathname, O_RDWR)) < 0)
	fatal("couldn't open %s: %m\n", pathname);

    char* orig_buf    = New char[HDR_SPACE + (LINUX_ALIGN - 1)];
    char* aligned_buf = align_up(orig_buf, LINUX_ALIGN);

    int rc;
    if ((rc = pread(fd, aligned_buf, HDR_SPACE, 0)) == 0 || rc < 0) 
	fatal ("couldn't read %d bytes from %s: %m\n", HDR_SPACE, pathname);
    else
	assert(rc == HDR_SPACE);

    daily_header* hdr = reinterpret_cast<daily_header*>(aligned_buf);

    off_t off2 = second_half_begins(hdr->max_blocks);

    char* zerobuf = New char[HDR_SPACE + (LINUX_ALIGN - 1)];
    char* zerobuf_a = align_up(zerobuf, LINUX_ALIGN);

    memset(zerobuf_a, 0, HDR_SPACE);
    if ((rc = pwrite(fd, zerobuf_a, HDR_SPACE, 0)) != HDR_SPACE)
	fatal("couldn't write %d bytes to %s at offset %lld\n",
	      HDR_SPACE, pathname, 0LL);
   
    warnx ("first half cleared\n");

    if ((rc = pwrite(fd, zerobuf_a, HDR_SPACE, off2)) != HDR_SPACE)
	fatal("couldn't write %d bytes to %s at offset %lld\n",
	      HDR_SPACE, pathname, off2);

    warnx ("second half cleared\n");

    delete [] orig_buf;
    delete [] zerobuf;

    close(fd);
}

void
ondisk_storage::init_async_iface()
{
    bool* b = New bool(false);

    a->open(fname, FILE_FLAGS, 0644,
	      wrap(this, &ondisk_storage::aopen_cb, b));

    while (!*b) 
	acheck();

    delete b;
}

void
ondisk_storage::aopen_cb(bool* b, ptr<aiofh> fh_ret, int err)
{
    if (fh_ret == NULL) 
	fatal("couldn't open file %s\n", fname.cstr());

    fh = fh_ret;
    *b = true;

    // drop privileges
    if (getuid() != geteuid())
	if (setuid(getuid()) < 0)
	    fatal("setuid(getuid()) failed: %m\n");

    if (getgid() != getegid())
	if (setgid(getgid()) < 0)
	    fatal("setgid(getgid()) failed: %m\n");

}


ondisk_storage::ondisk_storage(const char* fname, ptr<base_logger> logger,
                               int32_t read_size, int32_t write_size,
			       uint32_t nindex_entries)
 :  fh(NULL), 
    fd(-1),
    a(NULL),
    fname(fname),
    readq_size(0),
    map(New index_map(nindex_entries)),
    yest_map(New index_map(nindex_entries)), 
    nindex_entries(nindex_entries),
    pending_writes(MAX_QUEUED_WRITES),
    state(INITTING),
    curr_reads_out(0),
    curr_writes_out(0),
    promised_writes(0),
    state_timer(NULL),
    exchange_timer(NULL),
    logger(logger),
    stopped(false),
    false_positives(0),
    exchange_pending(false)
{    
    init_disk_constants(read_size, write_size);
    warnx("read: %d. write: %d\n", BLOCK_SIZE, FLUSH_SIZE);

    a = New aiod(naiod_procs, AIOD_SHM_SIZE, AIOD_SHM_MAXBUF);

    init_keyval_store();
    init_async_iface();

    state = DOING_WRITES;
    state_timer = delaycb(WRITE_INTERVAL_s, WRITE_INTERVAL_ns, 
                           wrap(this, &ondisk_storage::switch_state,
			              SETTLING_WRITES));

    schedule_next_exchange();
}

ondisk_storage::~ondisk_storage()
{
    if (!stopped) stop();
    delete map;
    delete yest_map;
}

void
ondisk_storage::stop()
{
    if (stopped) return;

    if (fh) {
	fh->close(cbi_null);
    }

    if (fd != -1)
	close(fd);

    timecb_remove(state_timer);
    timecb_remove(exchange_timer);

    state_timer  = NULL;
    exchange_timer = NULL;

    // this call should prevent any outstanding requests from finishing
    a->finalize();
    stopped = true;
    warnx("ondisk_storage stopping. false_positives: %lld\n", false_positives);
}


epoch_t
curr_epoch()
{
    struct tm* tm_p;
    time_t t;
    epoch_t ret;

    time(&t);
    tm_p = localtime(&t);

    ret.year  = tm_p->tm_year + 1900;
    ret.yday  = tm_p->tm_yday;

    return ret;
}

epoch_t
prev_epoch()
{
    struct tm* tm_p;
    time_t t;
    epoch_t ret;
    
    time(&t);
    tm_p = localtime(&t);

    if (tm_p->tm_mon == 0 && tm_p->tm_mday == 1) {
	// current day is new year's day, so previous day is new year's
	// eve
	struct tm ny_eve; 
	memset(&ny_eve, 0, sizeof(ny_eve));
	ny_eve.tm_year = tm_p->tm_year - 1;
	ny_eve.tm_mon  = 11;
	ny_eve.tm_mday = 31;

	time_t ny_eve_tt = mktime(&ny_eve);
	if (ny_eve_tt == (time_t)(-1))
	    fatal("mktime failed in function prev_epoch\n");

	struct tm* ny_eve_p = localtime(&ny_eve_tt);

	ret.year = ny_eve_p->tm_year + 1900;
	ret.yday = ny_eve_p->tm_yday;

    } else {
	ret.year = tm_p->tm_year + 1900;
	ret.yday = tm_p->tm_yday - 1;
    }
    return ret;
}

inline char*
align_up(char* p, ssize_t align)
{
    intptr_t p_int = reinterpret_cast<intptr_t>(p);
    intptr_t algn_p_int = (p_int + (align - 1)) & ~(align - 1);
    return reinterpret_cast<char*>(algn_p_int);
}

bool
aligned(char* p, size_t align)
{
    return !(reinterpret_cast<uintptr_t>(p) & (align - 1));
}

void
write_header(int fd, str fname, off_t offset, struct daily_header* h)
{
    char* orig_buf    = New char[HDR_SPACE + (LINUX_ALIGN - 1)];
    char* aligned_buf = align_up(orig_buf, LINUX_ALIGN);

    memset(aligned_buf, 0, HDR_SPACE);
    memcpy(aligned_buf, h, sizeof(struct daily_header));

    if (pwrite(fd, aligned_buf, HDR_SPACE, offset) != HDR_SPACE)
	fatal("couldn't write %d bytes to %s at offset %lld\n", 
	      HDR_SPACE, fname.cstr(), offset);

    delete [] orig_buf;
}

void
setup_half(int fd, str fname, off_t offset)
{
    struct daily_header h;

    h.dqm_magic_num = DQM_MAGIC_NUM;
    h.num_blocks    = 0;
    h.max_blocks    = MAX_BLOCKS_DAILY;
    h.epoch         = curr_epoch();

    write_header(fd, fname, offset, &h);
}

// returns false if the read could not be completed b/c at EOF, true
// otherwise.
bool
read_header(int fd, str fname, daily_header* hdr, off_t off)
{
    char* orig_buf    = New char[HDR_SPACE + (LINUX_ALIGN - 1)];
    char* aligned_buf = align_up(orig_buf, LINUX_ALIGN);

    int rc;
    ssize_t hdr_size = sizeof(*hdr);

    if ( (rc = pread(fd, aligned_buf, HDR_SPACE, off)) == 0) {
	delete [] orig_buf;
	return false;
    } else if (rc < 0) 
	fatal("couldn't read %d bytes from %s: %m\n", hdr_size, fname.cstr());
    else
	assert(rc == HDR_SPACE);

    memcpy(static_cast<void*>(hdr), aligned_buf, hdr_size);
    delete [] orig_buf;

    return true;
}

bool
valid_dqm_file(int fd, str fname)
{
    struct stat st;
    struct daily_header hdr;

    if (fstat(fd, &st) < 0)
	fatal("couldn't stat %s: %m\n", fname.cstr());

    ssize_t hdr_size = sizeof(daily_header);

    // unless it's a block file, make sure the file is bigger than the
    // length of the daily header.
    if (!S_ISBLK(st.st_mode) && st.st_size < hdr_size) 
	return false;

    read_header(fd, fname, &hdr, 0);

    return KEYVAL_HDR_VALID(hdr);
}

void
validate_fname(const char* fname)
{
    int rc;
    struct stat st;

    if (!strncmp("/dev/", fname, 5)) {

	// caller wants to open a device. we need to make sure it exists
	// and that it's a block device.
	if ((rc = stat(fname, &st)) != 0) 
	    fatal("couldn't open %s: %m\n", fname);
	else if (!S_ISBLK(st.st_mode)) 
	    fatal("fname %s must be either block device or file\n", fname);
	return;
    }
    
    const char* delim;
    if ( (delim = strrchr(fname, PATH_DELIM)) == NULL) 
	// no dirs
	return;

    int len          = delim - fname;
    char* const dir  = New char[len + 1];
    strncpy(dir, fname, len);
    dir[len] = '\0';

    // if stat of dir portion of filename fails, try to create the
    // directory portion. if stat succeeds, keep going.
    if ((rc = stat(dir, &st)) != 0) {   
	if (errno != ENOENT)
	    fatal("stat of dir %s failed: %m\n", dir);
	else if (!makedirs(dir))
	    fatal("dir %s didn't exist. couldn't create it: %m\n", dir);
    }

    delete [] dir;
}

inline off_t
second_half_begins(index_t max_blocks)
{
    return (HDR_SPACE + static_cast<off_t>(max_blocks)*BLOCK_SIZE);
}

void
ondisk_storage::init_offsets(off_t off2_beg, half_t curr_half)
{
    if (curr_half == FIRST_HALF) {
	today_start_off = 0;
	yest_start_off  = off2_beg;
    } else {
	today_start_off = off2_beg;
	yest_start_off  = 0;
    }

    curr_write_off  = today_start_off + HDR_SPACE + 
                       static_cast<off_t>(curr_blk)*BLOCK_SIZE;
}

void
ondisk_storage::init_keyval_store()
{
    struct daily_header hdr, hdr2;
    bool hdr2_valid = false;
    bool read_succeeded;

    validate_fname(fname.cstr());

    if ((fd = open(fname, O_CREAT | FILE_FLAGS, 0644)) < 0)
	fatal("could not open %s: %m\n", fname.cstr());

    if (!valid_dqm_file(fd, fname)) 
	setup_half(fd, fname, 0);
 
    // When we get to here, we have a file that is a valid DQM key-value
    // store. What we don't know is whether the *second half* of the
    // file is valid, and we also don't know whether either half is
    // "current" (i.e., whether either half represents the current or
    // past epoch). The purpose of the code below is to recover any
    // "half" that corresponds to the current or previous epoch.
   
    epoch = curr_epoch();
    epoch_t pre_epoch = prev_epoch();
    curr_blk = 0;
    read_succeeded = read_header(fd, fname, &hdr, 0);
    assert(read_succeeded); // no reason for the read to fail


    if (hdr.epoch == epoch) {
	warnx("recovering %d indices for today in first half\n",
	      static_cast<u_int32_t>(hdr.num_blocks));
	// first half corresponds to current epoch. recover the indices.
	recover_indices(fd, HDR_SPACE, hdr.num_blocks, map);
	curr_blk = hdr.num_blocks;
    } else if (hdr.epoch == pre_epoch) {
	warnx("recovering %d indices for yesterday in first half\n",
	      static_cast<u_int32_t>(hdr.num_blocks));
	// first half corresponds to previous epoch. recover the indices
	// and put them in the map for the previous epoch.
	recover_indices(fd, HDR_SPACE, hdr.num_blocks, yest_map);
    }

    max_blocks = hdr.max_blocks;
    off_2      = second_half_begins(max_blocks);
    read_succeeded = read_header(fd, fname, &hdr2, off_2);

    if (read_succeeded && KEYVAL_HDR_VALID(hdr2)) {
	// second half is a valid dqm key-value store
	hdr2_valid = true;
	if (hdr2.epoch == epoch) {
	    warnx("recovering %d indices for today in second half\n",
	          static_cast<u_int32_t>(hdr2.num_blocks));
	    // second half corresponds to current epoch. again, recover
	    // the indices.
	    recover_indices(fd, off_2 + HDR_SPACE, hdr2.num_blocks, map);
	    // we have to check that hdr.epoch != epoch b/c if both the
	    // first half and the second half refer to the current epoch
	    // (b/c they were both initialized) then we'll operate out
	    // of the first half, and curr_blk was set above.
	    if (hdr.epoch != epoch)
		curr_blk = hdr2.num_blocks;
	} else if (hdr2.epoch == pre_epoch) {
	    warnx("recovering %d indices for yesterday in second half\n",
		  static_cast<u_int32_t>(hdr2.num_blocks));
	    // second half corresponds to previous epoch. again, recover
	    // the indices and put them in the map for the previous
	    // epoch.
	    recover_indices(fd, off_2 + HDR_SPACE, hdr2.num_blocks, yest_map);
	}
    } else {
	setup_half(fd, fname, off_2);
	read_header(fd, fname, &hdr2, off_2);
    }

    half_t half;

    // we start in the second half if either 
    // (a) the first half corresponds to the previous epoch
    // (b) the second half is current *and* the first half is not current
    // otherwise, the first half is current or it's not current, but
    // then neither is the second half, in which case we should start in
    // the first half.
    half = ((hdr.epoch == pre_epoch) ||
	    (hdr2_valid && hdr2.epoch == epoch && hdr.epoch != epoch))
	     ? SECOND_HALF
	     : FIRST_HALF;

    strbuf sb("current half: %d. curr_blk: %d\n", 
              half, static_cast<u_int32_t>(curr_blk));
    logger->log(str(sb).cstr());

    init_offsets(off_2, half);
}

// Reads in 'num_blocks' blocks in file descriptor 'fd', starting 
// at offset 'start'. records the indices in the file of each key value
// pair in 'map'.
void
ondisk_storage::recover_indices(int fd, off_t start, 
                                index_t num_blocks, index_map* map)
{
    // Use FLUSH_SIZE because that will give us good throughput since
    // it's probably the track size.  Also, Linux requires that user
    // buffer be 512 byte aligned when writing to files opened with
    // O_DIRECT, so we align the buffer.
    char* orig_buf     = New char[FLUSH_SIZE + (LINUX_ALIGN - 1)];
    char* aligned_buf  = align_up(orig_buf, LINUX_ALIGN);  

    off_t    end = start + static_cast<off_t>(num_blocks) * BLOCK_SIZE;
    index_t  blk = 0;
    off_t    off;
    ssize_t  nbytes;
    const kv_block* block;
    ptr<dqm_key> key;

    for (off = start; off < end; off += nbytes) {

	nbytes = (off + FLUSH_SIZE <= end) ? FLUSH_SIZE : end - off;

	assert((nbytes % BLOCK_SIZE) == 0);

	if (pread(fd, aligned_buf, nbytes, off) != nbytes)
	    fatal("couldn't read %d bytes from %s at offset %lld: %m\n", 
		  nbytes, fname.cstr(), off);
	int num_blocks  = nbytes / BLOCK_SIZE;

	// loop through all of the blocks in the buffer we just read in
	for (int i = 0; i < num_blocks; i++) {
 
	    block = reinterpret_cast<const kv_block*>
	                         (aligned_buf + i*BLOCK_SIZE);

	    // loop through all of the key-value pairs in the current block
	    for (u_int32_t pair = 0; pair < PAIRS_PER_BLOCK; pair++) {

		key = New refcounted<dqm_key>();
		memcpy(key->base(), block->pairs[pair].key, HASH_SIZE);
		map->insert(key, blk);
	    }

	    blk++;
	}
    }

    delete [] orig_buf;

    assert(blk == num_blocks);
}

void
ondisk_storage::exchanger()
{
    assert(!exchange_pending);

    exchange_pending = true;

    warnx << "exchange_pending is true\n";
}

void
ondisk_storage::exchange_worker()
{
    warnx << "about to do flush work\n";

    // write the header so we store a complete record of all of our
    // work. this step is probably optional, given the flush_header
    // call.
    struct daily_header hdr;
    hdr.dqm_magic_num = DQM_MAGIC_NUM;
    hdr.num_blocks    = curr_blk;
    hdr.max_blocks    = max_blocks;
    hdr.epoch         = epoch;
    write_header(fd, fname, today_start_off, &hdr);
   
    // switch all variables around. note that curr_epoch() now gives the
    // current day, i.e., the newly born epoch in which we're running
    epoch           = curr_epoch(); 
    curr_blk        = 0;
    yest_start_off  = today_start_off;
    today_start_off = (today_start_off == 0) ? off_2 : 0;
    curr_write_off  = today_start_off + HDR_SPACE;

    delete yest_map;
    yest_map = map;
    map = New index_map(nindex_entries);

    schedule_next_exchange();
}

// schedule the next exchange for midnight in in the next epoch
void
ondisk_storage::schedule_next_exchange()
{
    time_t tt_curr;
    time(&tt_curr);

    struct tm* tm_curr;
    tm_curr = localtime(&tt_curr);

    struct tm tm_nxt;
    memset(&tm_nxt, 0, sizeof(tm_nxt));

    if (tm_curr->tm_mon == 11 && tm_curr->tm_mday == 31) {
	tm_nxt.tm_year = tm_curr->tm_year + 1;
	tm_nxt.tm_mon  = 0;
	tm_nxt.tm_mday = 1;
    } else {
	tm_nxt.tm_year = tm_curr->tm_year;
	tm_nxt.tm_mon  = tm_curr->tm_mon; // November 31, e.g., is allowed
	tm_nxt.tm_mday = tm_curr->tm_mday + 1;
#if 0
	// for debugging -- says that we schedule a switcher every
	// minute.
	tm_nxt.tm_mday = tm_curr->tm_mday;
	tm_nxt.tm_hour = tm_curr->tm_hour;
	tm_nxt.tm_min  = tm_curr->tm_min + 1;
#endif
    }

    time_t tt_nxt = mktime(&tm_nxt);
    assert(tt_nxt != (time_t)-1);

    struct timespec ts_nxt;
    ts_nxt.tv_sec = tt_nxt;
    ts_nxt.tv_nsec = 0;

    exchange_timer = timecb(ts_nxt, wrap(this, &ondisk_storage::exchanger));
}

// Give a quick answer. if 'key' is not in pending writes, then it may
// or may not be on the disk. We can tell by checking the two data
// structures (yesterday's and today's) that map keys to offsets.
void 
ondisk_storage::get(ptr<dqm_key> key, getcb cb)               
{
    if (kv_rec* rec = pending_writes.lookup(key, false)) {

	// the write has been requested but not yet written 
	(*cb)(rec->val);
	return;

    } 
   
    // 'key' may or may not be on disk, and it might be in the
    // current or previous index map. first check the current index map.
    index_t idx = map->lookup(key);

    if (idx == index_map::EMPTY_ENTRY)  
	try_yest(key, cb);
    else {
	// one_read_3() will figure out if this is a false positive
	diskread(idx, today_start_off, key, 
	                 wrap(this, &ondisk_storage::get_2, key, cb));
    }
    check_state();
}

// when we get to this function, the 'disk' has called us back to tell
// us whether the idx in the 'today map' really did point to the
// key,value pair we were looking for. if so, callback the original
// caller. if not, check yesterday's map.
void
ondisk_storage::get_2(ptr<dqm_key> key, getcb cb, ptr<dqm_val> val)
{
    if (val != NULL)
	(*cb)(val);
    else
	try_yest(key, cb);

    // to keep things moving, we see whether reads or writes can happen.
    check_state();
}


void
ondisk_storage::try_yest(ptr<dqm_key> key, getcb cb)
{
    index_t idx = yest_map->lookup(key);

    if (idx == index_map::EMPTY_ENTRY)
	(*cb)(NULL);
    else 
	diskread(idx, yest_start_off, key, cb);
    // no need to call check_state() here because both of the functions
    // that call this one call check_state()
} 

void
ondisk_storage::diskread(index_t idx, off_t base, ptr<dqm_key> key, getcb cb)
{
    // begin the read at the block containing the key we're looking for. 
    off_t off = base + HDR_SPACE + static_cast<off_t>(idx)*BLOCK_SIZE;
    readq.insert_tail(New read_task(key, cb, off));
    ++readq_size;

    if (readq_size > MAX_QUEUED_READS) {

	read_task* curr = readq.remove(readq.first);
	delete curr;
	--readq_size;
	assert(readq_size == MAX_QUEUED_READS);
    }

    // sanity check
    if (state == DOING_READS && curr_reads_out < MAX_READS_OUT) 
	assert(readq_size == 1);

    check_state();
}

void
ondisk_storage::check_reads()
{
    assert(state == DOING_READS);

    // Try as many reads as we have room for in our "window".
    u_int32_t lim = min(readq_size, MAX_READS_OUT - curr_reads_out);

    for (u_int32_t i = 0; i < lim; i++) 
	one_read();
}

void
ondisk_storage::one_read()
{
    read_task* curr = readq.first;
    readq.remove(curr);
    --readq_size;

    ++curr_reads_out;
    abufalloc(BLOCK_SIZE,	
	    wrap(mkref(this), &ondisk_storage::one_read_2, curr)); 
}

void
ondisk_storage::one_read_2(read_task* task, ref<aiobuf> abuf)
{
    assert(aligned(abuf->base(), LINUX_ALIGN));
    off_t off = task->off;
    fh->read(off, abuf,
	     wrap(mkref(this), &ondisk_storage::one_read_3, task));
}

void
ondisk_storage::one_read_3(read_task* task,
		          ptr<aiobuf> abuf, ssize_t n, int err)
{
    if (n != BLOCK_SIZE) {

	strbuf sb("one_read_3 failed. n: %d. err: %s", n, strerror(err));
	logger->log(str(sb).cstr());
	
	(*task->cb)(NULL);

    } else {

	const kv_block* block = reinterpret_cast<const kv_block*>
	                                (abuf->base());
	const char* keybase = task->key->base();
	u_int32_t pair;

	// Loop through the whole block looking for the key requested by
	// the read_task (though note that the key may not actually be
	// in the block).
	for (pair = 0; pair < PAIRS_PER_BLOCK; pair++) {

	    if (!memcmp(keybase, block->pairs[pair].key, HASH_SIZE)) {

		ptr<dqm_val> val = New refcounted<dqm_val>();
		memcpy(val->base(), block->pairs[pair].val, HASH_SIZE);
	                               
		// finally, the original caller gets called back
		(*task->cb)(val);
		break;
	    }
	}
	
	// false positive. key wasn't on disk after all
	if (pair == PAIRS_PER_BLOCK) {
	    ++false_positives;
	    (*task->cb)(NULL);
	}
    }

    delete task;
    --curr_reads_out;

    check_state();
}

// we copy the arguments to this function instead of using the global
// copies because this read could be happening near an 'exchange' boundary,
// and we want to the write to the "half" that was current at the time
// we invoked the flush operation.
void
ondisk_storage::flush_header(index_t idx, index_t max_blks, epoch_t ep,
                             off_t off)
{
    abufalloc(HDR_SPACE, 
	      wrap(mkref(this), &ondisk_storage::flush_header_2,
	           idx, max_blks, ep, off));
}

void
ondisk_storage::flush_header_2(index_t idx, 
			       index_t max_blks, 
			       epoch_t ep,
			       off_t   off,
			       ref<aiobuf> abuf)
{
    assert(aligned(abuf->base(), LINUX_ALIGN));

    daily_header hdr;
    hdr.dqm_magic_num = DQM_MAGIC_NUM;
    hdr.num_blocks    = idx;
    hdr.max_blocks    = max_blks;
    hdr.epoch         = ep;

    memset(abuf->base(), 0, HDR_SPACE);
    memcpy(abuf->base(), &hdr, sizeof(hdr));

    fh->write(off, abuf, wrap(mkref(this), &ondisk_storage::flush_header_3));

}

void
ondisk_storage::flush_header_3(ptr<aiobuf> abuf, ssize_t n, int err)
{
    if (n == -1) {
	strbuf sb("error in flush_header. err: %s", strerror(err));
	logger->log(str(sb).cstr());
	return;
    }

    assert(n == HDR_SPACE);
}

void 
ondisk_storage::put(ptr<dqm_key> key, ptr<dqm_val> val)          
{
    // Don't write key if it's already in the pending_writes.
    if (pending_writes.lookup(key, false) != NULL)
	return;

    index_t idx = map->lookup(key);

    if (idx == index_map::EMPTY_ENTRY) 
	// key is definitely novel; queue it up for writing.
	pending_writes.add(key, val);
    else 
	// Maybe we've seen this key before. Need to do a disk read.
	diskread(idx, today_start_off, key, 
	        wrap(this, &ondisk_storage::put_2, key, val));

    check_state();
}

// this is called when we're doing a put(), and there is a chance that
// the put() is a duplicate.
void
ondisk_storage::put_2(ptr<dqm_key> key, ptr<dqm_val> orig_val,
                      ptr<dqm_val> ret_val)
{

    if (ret_val != NULL) {
	// found key, so we don't insert into our in-memory stash
	static equals<ptr<dqm_key> > eq;
	assert( eq(orig_val, ret_val) );
	return;
    }

    // Key could have found its way into the pending writes while we
    // were out to lunch looking at the disk.
    if (pending_writes.lookup(key, false) != NULL)
	return;

    // Race condition. There could be a write request outstanding that
    // is going to put the key on the disk. This request could complete
    // just after the current function exits, which would have the
    // effect of adding (key,val) to the overflow table possible twice.
    // The pending_writes.add() that we do here will cause (key,val) to
    // go in the overflow table, and the key being written to the disk
    // is presumably reflected in the hash table's byte array or its
    // overflow table.
    pending_writes.add(key, orig_val);

    check_state();
}

void
ondisk_storage::check_writes()
{
    assert(state == DOING_WRITES);

    // the first condition just says, "taking into account that you've
    // already promised yourself a few writes, do we have enough writes
    // pending to do another write, since we always flush writes in a
    // group of FLUSH_NUM at a time"
    while (pending_writes.size() >= (promised_writes + 1)*PAIRS_PER_FLUSH &&
           curr_writes_out < MAX_WRITES_OUT &&
           curr_blk < max_blocks) 
	     one_write();
}

void
ondisk_storage::one_write()
{
    ++curr_writes_out;
    ++promised_writes;
    abufalloc(FLUSH_SIZE, 
	      wrap(mkref(this), &ondisk_storage::one_write_2, 
	           curr_blk, curr_write_off));

    curr_blk       += BLOCKS_PER_FLUSH;
    curr_write_off += FLUSH_SIZE;
}


// Need to pass in 'idx' and 'off' because there may be multiple
// 'promised' reads outstanding, and we have to tell this function which
// 'area' of the disk it gets to write to.
void
ondisk_storage::one_write_2(index_t idx, off_t off, ref<aiobuf> abuf)
{
    assert(aligned(abuf->base(), LINUX_ALIGN));

    u_int32_t blk_num, pair;
    char* base = abuf->base();
    struct kv_block* block;
    struct kv_rec*   rec;

    assert(pending_writes.size() >= PAIRS_PER_FLUSH);
 
    // Several blocks form a single write (this is to get good write
    // performance).
    for (blk_num = 0; blk_num < BLOCKS_PER_FLUSH; blk_num++) { 

	block = reinterpret_cast<kv_block*>(base + BLOCK_SIZE*blk_num);	

	for (pair = 0; pair < PAIRS_PER_BLOCK; pair++) {

	    rec = pending_writes.oldest();
	    memcpy(block->pairs[pair].key, rec->key->base(), HASH_SIZE);
	    memcpy(block->pairs[pair].val, rec->val->base(), HASH_SIZE);

	    map->insert(rec->key, idx);
	    pending_writes.remove(rec);
	}

	idx++;
    }

    fh->write(off, abuf, wrap(mkref(this), &ondisk_storage::one_write_3));

    --promised_writes;
}

void
ondisk_storage::one_write_3(ptr<aiobuf> abuf, ssize_t n, int err)
{
    --curr_writes_out;
    if (n == -1) {
	strbuf sb("error in one_write_3. err: %s", strerror(err));
	logger->log(str(sb).cstr());
	return;
    }
   
    assert(n == FLUSH_SIZE);

    check_state();
}


void
ondisk_storage::abufalloc(int size, callback<void, ref<aiobuf> >::ref cb)
{
    if (ptr<aiobuf> b = a->bufalloc(size))
	(*cb)(b);
    else
	a->bufwait(wrap(mkref(this), &ondisk_storage::abufalloc, size, cb));
}

void
ondisk_storage::switch_state(state_t new_state)
{
    state_timer = NULL;

#ifdef PRINT_SETTLE_TIMES
    const char* m = (new_state == SETTLING_READS) 
		      ? "SETTLING_READS" 
		      : "SETTLING_WRITES";
    char foobuf[256];
    sprintf(foobuf, "entering %s:\t%.3f\n", m, 
		     (double)get_now() / 1000.);
    warnx("%s", foobuf);
#endif
    state = new_state;
    check_state();
}

void 
ondisk_storage::check_state()
{
    switch (state) {

    case DOING_READS:

	check_reads();
	break;

    case DOING_WRITES:

	check_writes();
	break;

    case SETTLING_READS:

	if (curr_reads_out == 0) {

#ifdef PRINT_SETTLE_TIMES
	    char foobuf[256];
	    sprintf(foobuf, "entering WRITES:\t\t%.3f\n\n", 
			     (double)get_now() / 1000.);
	    warnx("%s", foobuf);
#endif

	    if (exchange_pending) {
		exchange_worker();
		exchange_pending = false;
	    }

	    state = DOING_WRITES;

	    state_timer = delaycb(WRITE_INTERVAL_s, WRITE_INTERVAL_ns,
				 wrap(this, &ondisk_storage::switch_state, 
					    SETTLING_WRITES));
	    check_writes();
	}
	break;

    case SETTLING_WRITES:

	if (curr_writes_out == 0) {

#ifdef PRINT_SETTLE_TIMES 
	    char foobuf[256];
	    sprintf(foobuf, "entering READS:\t\t\t%.3f\n\n", 
			    (double)get_now() / 1000.);
	    warnx("%s", foobuf);
#endif
	    assert(promised_writes == 0);

	    // Begin reading. Go through the list of queued reads and,
	    // one-by-one read the (key, value) pair from the disk. As
	    // an opportunistic hack, we flush the header to the
	    // beginning of the key-value store (since we're random
	    // accessing the disk at this point anyway).

	    state = DOING_READS;
	    state_timer = delaycb(READ_INTERVAL_s, READ_INTERVAL_ns,
				  wrap(this, &ondisk_storage::switch_state, 
					     SETTLING_READS));
	    flush_header(curr_blk, max_blocks, epoch, today_start_off);
	    check_reads();
	}
	break;


    default:
	assert(false);
    }

}


