/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#ifndef _KEYVAL_STORE_H_
#define _KEYVAL_STORE_H_

#include "async.h"
#include "dqm_node.h"
#include "list.h"
#include "lru_cache.h"
#include "index_map.h"

class aiofh;
class aiobuf;
class aiod;
class base_logger;

struct epoch_t { 
    u_int32_t year; 
    u_int32_t yday;
};
inline bool operator==(epoch_t& e1, epoch_t& e2) {
    return (e1.yday == e2.yday && e1.year == e2.year);
}
inline bool operator!=(epoch_t& e1, epoch_t& e2) {
    return (e1.yday != e2.yday || e1.year != e2.year);
}

// utility functions for use by other programs
void init_disk_constants(int32_t read_size = -1, int32_t write_size = -1);
void clear_storage(char* pathname);

class ondisk_storage : public virtual refcount {

    public:

	// if read_size or write_size is -1, this object will use
	// default values
	ondisk_storage(const char* fname, ptr<base_logger>, 
	               int32_t read_size, int32_t write_size,
		       uint32_t nindex_entries = 10000000);
	~ondisk_storage();

	void stop();

	void put(ptr<dqm_key> k, ptr<dqm_val> v);
	void get(ptr<dqm_key>, getcb);

	// which_half should be 0 or 1. This call clears the given half.
	void clear(int which_half);

    private:

	typedef enum {FIRST_HALF, SECOND_HALF} half_t;

	// Keep track of whether we're reading, writing, waiting for
	// reads to 'settle' (i.e., for the outstanding reads to
	// complete), or waiting for writes to 'settle'.
	typedef enum {INITTING, DOING_WRITES, SETTLING_WRITES, 
	              DOING_READS, SETTLING_READS} state_t;

	struct read_task {
	    read_task(ptr<dqm_key> k, getcb cb, off_t off) 
	        : key(k), cb(cb), off(off) {}
	    ptr<dqm_key> key;
	    getcb cb;
	    off_t off;
	    tailq_entry<read_task> qlink;

	    static int cmp(const void* p1, const void* p2) {
		const read_task*const* t1 = 
		      static_cast<const read_task* const *>(p1);
		const read_task*const* t2 = 
		      static_cast<const read_task* const *>(p2);
		u_int32_t o1 = (*t1)->off;
		u_int32_t o2 = (*t2)->off;
		if (o1 < o2) return -1;
		if (o1 > o2) return 1;
		return 0;
	    }
	};

	// functions to init
	void init_offsets(off_t off2_beg, half_t curr_half);
	void init_keyval_store();
	void init_async_iface();
	void aopen_cb(bool*, ptr<aiofh>, int);
	void recover_indices(int fd, off_t start, index_t num_pairs,
			     index_map* map);

	// glue
	void get_2(ptr<dqm_key>, getcb, ptr<dqm_val>);
	void put_2(ptr<dqm_key>, ptr<dqm_val> orig, ptr<dqm_val> ret);
	void try_yest(ptr<dqm_key>, getcb);

	// needed for both reads and writes
	void abufalloc(int size, callback<void, ref<aiobuf> >::ref cb);

	// functions to write
	void enqueue_diskwrite(ptr<dqm_key>, ptr<dqm_val>);
	void check_writes();
	void one_write();
	void one_write_2(index_t idx, off_t off, ref<aiobuf>);
	void one_write_3(ptr<aiobuf>, int n, int err);

	// functions to read
	void diskread(index_t idx, off_t base, ptr<dqm_key> key, getcb cb);
	void check_reads();
	void one_read();
	void one_read_2(read_task*, ref<aiobuf> abuf);
	void one_read_3(read_task*, ptr<aiobuf> abuf, ssize_t n, int err);

	// functions to write the header with the current size (for 
	// crash recovery)
	void flush_header(index_t idx, index_t max_blks, 
	                  epoch_t ep, off_t off);
	void flush_header_2(index_t idx, index_t max_prs, epoch_t ep,
                            off_t off, ref<aiobuf> abuf);
	void flush_header_3(ptr<aiobuf> abuf, ssize_t n, int err);

	// State machine for switching between the reading task and the
	// writing task
	void check_state();
	// Helper function.
	void switch_state(state_t new_state);

	// switch between the 'yesterday' half and the 'today' half
	void exchanger();
	void exchange_worker();
	void schedule_next_exchange();

	ptr<aiofh> fh;       // points to the file holding the keys and values
	int        fd;       // local file descriptor.
	aiod*      a;	     // our interface to the async I/O daemon
	str        fname;    // the name of the key-value file

	// state to manage the epoch change
	index_t    max_blocks;  // how many pairs we can hold in each half
	off_t      off_2;       // where the second "half" begins

	// state that changes in each epoch
	epoch_t    epoch;	    // current epoch
	index_t    curr_blk;	    // current index for writing
	off_t      today_start_off; // beg. of today's half, incl. header
	off_t      yest_start_off;  // beg. of yesterday's half, incl.  header
	off_t      curr_write_off;  // where the next write will happen

	u_int32_t  readq_size;	    // number of not-yet-serviced reads

	index_map*  map;
	// This map holds the indices into the key-value store from the
	// previous epoch. These indices later must be mapped to actual
	// offsets in the file.
	index_map*  yest_map;  

	uint32_t nindex_entries;

	// We use an LRU cache b/c we want the abstraction of a queue
	// and a map.
	lru_cache  pending_writes;

	// A queue of requests that will ultimately be sorted by offset
	// on the disk and then read from the disk
	tailq<read_task,  &read_task::qlink>   readq;

	state_t state;
	u_int32_t curr_reads_out;
	u_int32_t curr_writes_out;
	// keep track of how many writes we 'promised' to ourselves
	// between one_write() and one_write_2()
	u_int32_t promised_writes;  

	timecb_t*  state_timer;
	timecb_t*  exchange_timer;
	
	ptr<base_logger> logger;

	bool stopped;
	u_int64_t false_positives;

	// true if this object is trying to exchange the two halves, false
	// otherwise. used as a hacky signal from one part of the code
	// to the other.
	bool exchange_pending;
};

#endif
