/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
//
// $Id: heap.h,v 1.2 2006/05/19 13:17:03 mwalfish Exp $
//
#ifndef _HEAP_H_
#define _HEAP_H_

#include <string.h>
#include <stdlib.h>

// A heap that stores objects of type T, with the 'min' of the
// collection on top.  T needs to have:
//   -- a '<'  operator defined.  
//   -- a constructor that can be called with 0, like this: T(0)
//      (integers or pointers already have this.)
// The heap will grow as the number of events do.


// this implementation is mostly copied from CLR.
#define parent(i) (i>>1)
#define left(i)   (i<<1)
#define right(i)  ((i<<1) | 0x1)

template <typename T>
class Heap {

    public:
	// argument specifies initial size of heap.
	Heap(int num = 1000);
	~Heap();

	void insert(T); 

	// if queue is empty, returns T(0)
	T extract_min();
	// if queue is empty, returns T(0)
	T inspect_min();

	int size() { return m_size; }
	int max()  { return m_max; }

	// for testing and debugging; not really public.
	T* dump() { return m_array; }
	void debug();

    private:

	void heapify();

	int m_size;
	int m_max;
	T*  m_array;
};

template <typename T> 
Heap<T>::Heap(int num) :
    // we'll index our array from 1, from the pseudo-code in CLR.
    m_size(0), m_max(num), m_array( (T*)calloc(num+1,sizeof(T)) )
{
}

template <typename T>
Heap<T>::~Heap()
{
    free(m_array);
}

template <typename T>
inline void Heap<T>::insert(T to_insert)
{
    if (m_size == m_max) {
	
	// we're at the top of the heap, so allocate twice as much
	// as we have and copy everything over.
	T* new_array = (T*)calloc(m_max*2 + 1, sizeof(T));
	memcpy(new_array, m_array, (m_size+1)*sizeof(T));
	free(m_array);
	m_array = new_array;
	m_max *= 2;
    }
    
    int idx = ++m_size;

    while (idx > 1 && to_insert < m_array[parent(idx)]) {

	m_array[idx] = m_array[parent(idx)];
	idx = parent(idx);
    }

    m_array[idx] = to_insert;
}

template <typename T> 
inline T Heap<T>::extract_min()
{
    if (m_size < 1) 
	return T(0);

    T ret = m_array[1];

    m_array[1] = m_array[m_size];
    m_size--;

    heapify();

    return ret;
}

template <typename T> 
inline T Heap<T>::inspect_min()
{
    if (m_size < 1)
	return T(0);

    return m_array[1];
}

template <typename T> 
inline void Heap<T>::heapify()
{
    int i = 1;	    // the 1-indexed array
    while (1) {

	int l = left(i);
	int r = right(i);

	int smallest = i;
	if (l <= m_size && m_array[l] < m_array[i])
	    smallest = l;

	if (r <= m_size && m_array[r] < m_array[smallest])
	    smallest = r;
	
	if (smallest != i) {

	    T temp = m_array[smallest];
	    m_array[smallest] = m_array[i];
	    m_array[i] = temp;
	    i = smallest;

	} else 

	    break;

    }
}


template <typename T>
void Heap<T>::debug()
{
    int i = 1;
    int l, r;
    for (i = 1; i <= m_size; i++) {
	l = left(i);
	r = right(i);
	if (l <= m_size) {
	    if (m_array[i] > m_array[l]) {
		printf("whoah. i,l: %d,%d. vals: %d,%d\n", 
			i, l, m_array[i], m_array[l]);
		exit(0);
	    }
	}
	if (r <= m_size) {
	    if (m_array[i] > m_array[r]) {
		printf("whoah. i,r: %d,%d. vals: %d,%d\n", 
			i, r, m_array[i], m_array[r]);
		exit(0);
	    }
	}
    }
}

#endif
