/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
/*
 * $Id: send.c,v 1.59 2006/05/19 13:17:03 mwalfish Exp $
 *
 * Main sending loop for DQM nodes test program. For TESTs, all
 * work (i.e., selecting the <k,v>, deciding whether the <k,v> is
 * reused, selecting a portal) is done right before sending the RPC.
 * For SETs, some of its work is done when the SET is _scheduled_
 * (because the answers to the work are determined by the TEST
 * request) while some of the work is done right before sending the SET RPC.
 */
#include <openssl/sha.h>
#include <errno.h>
#include <map>
#include "reusemap.h"
#include "u.h"
#include "send.h"
#include "util2.h"

typedef struct _context {

    int  s;  // the socket

    portals* ps;

    // key data structures
    reuse_map_t reusemap;  /* keeps track of which is spam */
    events eventq;	   
    xidmap_t xidmap;	   /* maps xids to outstanding TEST reqs */

    uint32 msgc_curr;	   /* for non-spam we keep incrementing this */
    uint32  spam_beg;      /* lowest numbered counter that could be spam */
    uint32  spam_max;	   /* never issue a spam TEST for more than this */
    double  pr_spam;
    bool preloading_done; 
    bool sending_done;
    stamp_cntr* stamp_cntrs;

    double mean_interarr_us;  /* avg. microseconds between TESTs */

    uint64 last_time;	      /* last time */
    uint32 num_sent;	      /* for counter dumpage */
    unsigned long last_sent;  /* for counter dumpage */

    bool first_tmo;          /* for debugging */
    portal_req req;          /* we always make our requests here */

    struct config_params* params;

    int lfd;                 /* log file */
    FILE *rfile;             /* rate file */
    uint64 rtime;            /* rate update interval */
} context;

context ctx;
struct counters cntrs;

#define DUMP_SPAM_STAT(stat) printf("\t" #stat ": %lu\n", stat);

#define NUMBUCKETS 33

#define BUCKET(count) (count < NUMBUCKETS ? count : (NUMBUCKETS - 1))

/* 
 * JD, note that if you want to keep track of each stamp separately,
 * you can simply look at the various entries in the ctx.stamp_cntrs
 * array. Each entry tells you the attempted number of TESTs on the
 * stamp, the number of times the stamp was found, the number of times
 * the stamp was not found, and the number of times that the request
 * timed out.
 */
void
dump_spam_stats()
{
    uint32 i;
    uint32 tmo = 0;
    uint32 att = 0;
    uint32 found = 0;
    uint32 notfound = 0;
    stamp_cntr* cntr;

    uint32 reuses[NUMBUCKETS];
    memset(reuses, 0, sizeof(reuses));

    for (i = 0; i < ctx.params->universe_sz; i++) {
	cntr = &ctx.stamp_cntrs[i];
	tmo      += cntr->tmo;
	att      += cntr->att;
	found    += cntr->found;
	notfound += cntr->notfound;
	// only count if we attempted at least one. we don't count
	// timeout events because we're trying to measure the theory,
	// and the theory assumes the portal is up. for the experiments
	// we run, the only way a timeout could occur is if the portal
	// is down.
	if (cntr->att > 0)
	   reuses[BUCKET(cntr->notfound)]++; 
    }
    printf("spam stats:\n");

    DUMP_SPAM_STAT(att);
    DUMP_SPAM_STAT(found);
    DUMP_SPAM_STAT(notfound);
    DUMP_SPAM_STAT(tmo);
    for (i = 0; i < NUMBUCKETS; ++i) 
       printf("\t" "%lu reuses: %lu\n", i, reuses[i]);
      
}

#define DUMP_STAT(stat) printf("\t" #stat ": %lu\n", cntrs.stat);

void dump_stats()
{
    printf("stats for spam and non-spam:\n");
    DUMP_STAT(set_req)
    DUMP_STAT(set_resp)
    DUMP_STAT(test_req)
    DUMP_STAT(test_found)
    DUMP_STAT(test_notfound)
    DUMP_STAT(test_tmo)
    DUMP_STAT(test_resp_after_tmo)
    DUMP_STAT(hash_check_failed)
    dump_spam_stats();
}

#define FLUSH_BUF_SIZE 4096

void log_to_latency(char *buf, int len, bool flush) {
  // len had better be less than FLUSH_BUF_SIZE!!!

  if (ctx.lfd == -1)
    return;

  static char flush_buf[FLUSH_BUF_SIZE];
  static int pos = 0;
  
  if ((len + pos >= FLUSH_BUF_SIZE) || flush) {
    write(ctx.lfd, flush_buf, pos);
    pos = 0;
  }
  memcpy(&flush_buf[pos], buf, len);
  pos += len;
}

void process_latency(uint32 latency, uint64 now, 
		     bool found, bool spammy)
{   
    if (ctx.lfd == -1)
      return; // keep this here anyway, to save a bit
              // of processing...

    char buf[32];
    int len;

    sprintf(buf, "%llu %lu %c%c\n", now, (uint32)latency, 
	    (found ? 'Y' : 'N'), (spammy ? 'S' : 'F'));
    len = strlen(buf);
    log_to_latency(buf, len, false);
}

void print_and_exit() {
  printf("message counter: %lu\n", ctx.msgc_curr);
  printf("outstanding tests: %lu\n", ctx.eventq.size() - 2L);
  log_to_latency("", 0, true);
  dump_stats();
  exit(0);
}

void
init_sock(int* s)
{
    *s = socket(AF_INET, SOCK_DGRAM, 0);
    assert(*s >= 0);

    int n = (1 << 20); 

    // expand socket buffer
    if (setsockopt(*s, SOL_SOCKET, SO_RCVBUF, (char*)&n, sizeof(n)) < 0)
	PERROR("SO_RCVBUF\n");

    if (setsockopt(*s, SOL_SOCKET, SO_SNDBUF, (char*)&n, sizeof(n)) < 0)
	PERROR("SO_SNDBUF\n");

    struct sockaddr_in src_addr;
    memset(&src_addr, 0, sizeof(src_addr));
    src_addr.sin_family      = AF_INET;
    src_addr.sin_port        = htons(OUR_WELL_KNOWN_PORT);
    src_addr.sin_addr.s_addr = INADDR_ANY;
    if (bind(*s, (struct sockaddr*)&src_addr, sizeof(src_addr)))
	PERROR("could not bind to our well-known port+1");
}

void 
init_req(portal_req* req) 
{
    portal_req_hdr* hdr = &req->hdr;

    /* choice of xid doesn't matter much for us */
    hdr->xid         = htonl(unifRand(0, MAX_INT/4)); 
    hdr->type        = htonl(RPC_CALL);
    hdr->rpc_version = htonl(RPC_VERSION);
    hdr->prog        = htonl(PORTAL_PROG);
    hdr->prog_vers   = htonl(PORTAL_PROG_VER);
    bzero(hdr->auth_dummy, sizeof(hdr->auth_dummy));
}

void 
do_send(int req_len, struct sockaddr_in* addr, char* poor_name)
{
    int rc;

    rc = sendto(ctx.s, (char*)&ctx.req, req_len, 0, 
	       (const struct sockaddr*)addr, sizeof(*addr));

    if (rc < 0) {
		
	if (errno != EAGAIN)  
	  printf("sending failure. node: %s: %s\n", 
		 poor_name, strerror(errno)); 
	else 
	  printf("EAGAIN. node: %s\n", poor_name);
			
    } else 
	++ctx.num_sent;
}

/*
 * update ctx to reflect the current time's rate as given by the ratefile.
 */
void
rate_update() {
  char buf[13];
  double rate;

  if ((fgets(buf, sizeof(buf), ctx.rfile)) == NULL) {
    printf("unable to read from rate file (out of data?): %s\n", 
	   strerror(errno));
    exit(-1);
  }
  rate = ((double) strtol(buf, NULL, 0)) / ctx.rtime;

  printf("setting rate to %f\n",rate);

  ctx.mean_interarr_us = 1000000.0 / rate;
}


/*
 * Pack 'ctx.req' with all of the info necessary to send a TEST
 * RPC. Returns a pointer to the information that the corresponding 
 * SET call will need. 
 */
void
make_test(outstanding_test* test_out)
{
    char crud_buf[256];
    uint32 n_spam;
    stamp_cntr* stmp_cntr;

    portal_req_hdr* hdr = &ctx.req.hdr;

    hdr->xid  = htonl(ntohl(hdr->xid) + 1);
    hdr->proc = htonl(PORTAL_TEST_PROC);

    if (ctx.preloading_done && funifRand(0.0, 1.0) < ctx.pr_spam) {

	uint32 range_low = ctx.spam_beg;
	uint32 range_high = ctx.spam_max;
	n_spam = ctx.reusemap.get_spam(range_low, range_high);

	test_out->is_spam = true;
	test_out->n_spam = n_spam;
	stmp_cntr = &ctx.stamp_cntrs[n_spam - ctx.spam_beg];
	stmp_cntr->att++;

    } else {

	n_spam = ctx.msgc_curr;
	if (ctx.msgc_curr == ctx.spam_max) {
	    printf("finished testing the spam universe\n");
	    char *msg = ";==FT==\n";
	    log_to_latency(msg, strlen(msg), false);
	    // we give ourselves a buffer before we declare that
	    // pre-loading is done because, before we set a 'spam', we
	    // want to make sure that the last TEST made it to the
	    // DQE infrastructure.
	    event ev(get_now() + RPC_TIMEOUT*1000000, PRELOAD_DONE);
	    ctx.eventq.insert(ev_pair(ev.wh, ev));
	}
	
	ctx.msgc_curr++;

	test_out->is_spam = false;
    }

    sprintf(crud_buf, "%lu", n_spam);
  
    // if this condition is true, then there's no work left for u-2
    if (ctx.msgc_curr > ctx.params->msgc_max || ctx.reusemap.size() == 0) {
      stop_running();
    }
     
    SHA1((uint8*)crud_buf, strlen(crud_buf), test_out->val);
    SHA1(test_out->val, KEY_VAL_LEN, ctx.req.body.test.key);


    return;
}

void
stop_running() {
  if (ctx.params->done_delay) {
    ctx.sending_done = true;
    event ev(get_now() + RPC_TIMEOUT*1000000, DONE_RUNNING);
    ctx.eventq.insert(ev_pair(ev.wh, ev));
  } else {
    print_and_exit();
  }
} 

/* 
 * Returns the length of the request; this function knows the
 * length because this function allocates the dummy metastamp.
 */
inline int 
make_set(uint8* val)
{
    int ms_len, leftover, padding;
    portal_req_hdr* hdr = &ctx.req.hdr;
    set_req_body*   set = &ctx.req.body.set;

    hdr->xid  = htonl(ntohl(hdr->xid) + 1);
    hdr->proc = htonl(PORTAL_SET_PROC);

    /* copy the val buffer to the RPC struct */
    memcpy(set->val, val, KEY_VAL_LEN);
    /* and now get the key into the RPC struct by SHAing the value */
    SHA1(val, KEY_VAL_LEN, set->key);

    /* metastamps are just placeholders now */
    memset(set->metastamp, 'a', 4);
    ms_len = 4;

    /* 4-byte alignment */
    leftover = ms_len % 4;
    if (leftover) {
	padding = 4 - leftover;
	bzero(&set->metastamp[ms_len], padding);
    } else {
	padding = 0;
    }

    set->ms_len = htonl(ms_len);

    return sizeof(portal_req_hdr) +   /* header plus ... */
           sizeof(set_req_body)   -   /* ... body minus ... */
           METASTAMP_MAX_SIZE     +   /* ... default ms size plus ... */
	   ms_len                 +   /* ... the actual ms size plus ... */
	   sizeof(set->ms_len)    +   /* ... bytes to hold ms_len plus ...  */
	   padding;                   /* ... padding, if any */
}

#define COMPLAIN(x, y) { printf(x, y); assert(false); }

// stupid helper function that masks the ugliness of finding the
// TEST_TMO event that is at time 'when'
void 
find_testtmo(uint64 when, int xid, ev_it& it)
{
    // Loop to find the event we care about. The only reason we should
    // be looping is if multiple events share the same time
    it = ctx.eventq.find(when);
    assert (it != ctx.eventq.end());
    while (it->second.xid != xid && it->second.type != TEST_TMO) {
	assert (it->second.wh == when);
	++it;
	assert (it != ctx.eventq.end());
    }
}

inline void
process_reply_test(int xid, portal_resp_body* resp_body, bool found)
{
    outstanding_test* out_test = ctx.xidmap[xid];
    if (out_test == NULL) {
	++cntrs.test_resp_after_tmo;
	return;
    }

    ctx.xidmap.erase(xid);

    /* record latency */
    if (unifRand(0, LATENCY_SAMPLE) == 0) {
	uint64 now = get_now();
	uint32 latency = now - out_test->begin;
	process_latency(latency, now, found, out_test->is_spam);
    }

    /* check hash */
    if (found && memcmp(out_test->val, resp_body->val, KEY_VAL_LEN))  {
	++cntrs.hash_check_failed;
	printf("Warning! Hash checked failed in test reply\n");
    }

    /* pull the timeout from the event queue */
    ev_it it;
    find_testtmo(out_test->when, xid, it);

    ctx.eventq.erase(it);

    /* do the SET, if applicable */
    if (!found && !ctx.params->tests_only) {
	portal* poor = out_test->poor;
	int req_len  = make_set(out_test->val);
	do_send(req_len, &poor->addr, poor->name);
	++cntrs.set_req;
#ifdef CHATTY
	dump_req(&ev, &ctx.req, poor);
#endif
    }

    /* bookkeeping for spam */
    if (out_test->is_spam) {
	uint32 idx = out_test->n_spam - ctx.spam_beg;
	stamp_cntr* stmp_cntr = &ctx.stamp_cntrs[idx];
	if (found) stmp_cntr->found++;
	else       stmp_cntr->notfound++;
    }

    free(out_test);
}

inline void
process_reply(const char* buf, uint32 len)
{
    portal_resp_hdr* resp_hdr = (portal_resp_hdr*)buf;

    if (len < RESP_HDR_SIZE) {
	hexdump(buf, len);	
	COMPLAIN("bad RPC header. length: %lu\n", len)
    }

    if (ntohl(resp_hdr->type) != RPC_REPLY)
	COMPLAIN("got packet with resp type: %d\n", ntohl(resp_hdr->type))

    if (ntohl(resp_hdr->reply_stat) != RPC_MSG_ACCEPTED)
	COMPLAIN("got packet with reply stat: %d\n", 
		 ntohl(resp_hdr->reply_stat))

    if (ntohl(resp_hdr->accept_stat) != RPC_SUCCESS) {
	printf("got packet with accept stat: %d\n", 
	       ntohl(resp_hdr->accept_stat));
	return;
    }

#ifdef CHATTY
    pretty_print_resp(resp);
#endif

    if (len != TEST_RESP_FOUND_SIZE && len != SET_RESP_SIZE) {
	hexdump(buf, len);
	COMPLAIN("not a legit RPC resp. length: %lu\n", len);
    }

    portal_resp* resp           = (portal_resp*)buf;
    portal_resp_body* resp_body = &resp->body;

    int xid = resp_hdr->xid;

    /* 
     * The "parsing" of the reply message is a hack that relies on the fact
     * that we only have two message types and that the result of the RPC,
     * for both TEST and SET, shares the same first four bytes.
     */
    switch (ntohl(resp_body->res)) {

	case DQM_OK:

	    if (len == SET_RESP_SIZE)  
		++cntrs.set_resp;
	    else {
		assert(len == TEST_RESP_FOUND_SIZE);
		process_reply_test(xid, resp_body, true);
		++cntrs.test_found;
	    }
	    break;

	case DQM_NOT_FOUND:
	    process_reply_test(xid, resp_body, false);
	    ++cntrs.test_notfound;
	    break;

	case DQM_NO_MORE_STORAGE:
	    COMPLAIN("ran out of storage: %d\n", ntohl(resp_body->res));
	    assert(false);
	    break;

	case DQM_BAD_PARAM:
	    COMPLAIN("got DQM bad parameter: %d\n", ntohl(resp_body->res));
	    assert(false);
	    break;

	default:
	    COMPLAIN("got unrecognized DQM response code: %d\n",
	             ntohl(resp_body->res));
	    assert(false);

    }

}

inline void
dump_counters()
{
    uint64 curr         = get_now();
    uint64 elapsed_usec = curr - ctx.last_time;
    double rate = (ctx.num_sent - ctx.last_sent)*1000000. / elapsed_usec;

    printf("rate: %.3f\n", rate);

    ctx.last_time = curr;
    ctx.last_sent = ctx.num_sent;
}

inline void
schedule_dump()
{
    event ev(get_now() + DUMP_INTERVAL_SEC*1000000, DUMP);
    ctx.eventq.insert(ev_pair(ev.wh, ev));
}

inline void
schedule_test(uint64 last)
{
    double interarrival_usec = fExp(ctx.mean_interarr_us);

    // round off 
    int interarr_usec_int = (int)interarrival_usec;
    double frac = interarrival_usec - interarr_usec_int;
    if (frac >= .5) ++interarr_usec_int;

#ifdef CHATTY
    pretty_print(get_now(), last, interarr_usec_int, "test");
#endif
    event ev(last + interarr_usec_int, TEST);
    ctx.eventq.insert(ev_pair(ev.wh, ev));
}

void 
setup_context(config_params* params)
{
    init_sock(&ctx.s);

    ctx.ps        = params->ps;
    ctx.spam_beg  = params->msgc_begin;
    ctx.msgc_curr = params->msgc_begin;
    ctx.spam_max  = params->universe_sz + params->msgc_begin;
    ctx.pr_spam   = params->pr_spam;
    ctx.preloading_done = false;
    ctx.sending_done = false;
    ctx.stamp_cntrs = (stamp_cntr*)calloc(params->universe_sz, 
					  sizeof(stamp_cntr));
    
    ctx.reusemap.init(ctx.spam_beg, ctx.spam_max, params->repeat);

    ctx.mean_interarr_us = 1000000.0 / (params->preload_rate);

    ctx.last_time = get_now();
    ctx.num_sent  = 0;
    ctx.last_sent = 0;
    ctx.first_tmo = false;

    init_req(&ctx.req);

    ctx.params    = params;

    if (ctx.params->latency_file) {

	if ((ctx.lfd = open(ctx.params->latency_file, 
			    O_WRONLY | O_APPEND | O_CREAT | O_TRUNC,
			    0644)) < 0) {
	    printf("failed to open latency log file: %s\n", strerror(errno));
	    exit(-1);
	}
    } else {
	ctx.lfd = -1;
    }
    if (ctx.params->rate_file) {
      if ((ctx.rfile = fopen(ctx.params->rate_file, "r")) == NULL) {
	printf("failed to open rate file: %s\n", strerror(errno));
	exit(-1);
      }
      if (ctx.params->rate_file_o > 0) {
	char buf[13];
	uint32 i;
	
	for (i = 0; i < ctx.params->rate_file_o; ++i) {
	  if ((fgets(buf, sizeof(buf), ctx.rfile)) == NULL) {
	    printf("error finding index %lu in rate file: %s\n",
		   ctx.params->rate_file_o, strerror(errno));
	    exit(-1);
	  }
	}
      }
      ctx.rtime = ctx.params->rate_file_t;
    } else {
      ctx.rfile = NULL;
    }
}

void 
process_event(event& ev)
{
    outstanding_test* out_test = NULL;
    portal* poor;
    uint64 tmo;
    uint64 now;
    int xid;
    xid_it it;
    char *msg;

    switch (ev.type) {

	case TEST:

	    /* 
	     * we assume a renewal process for TESTs, so to schedule
	     * the next one we only need to know when the previous one
	     * happened.
	     */
	    if (ctx.sending_done) {
	      break;
	    }

	    schedule_test(ev.wh);
	    out_test = (outstanding_test*)malloc(sizeof(outstanding_test));
	    make_test(out_test);

	    /* 
	     * Record the portal so that the SET can go to the same
	     * portal as the TEST.
	     */
	    poor = ctx.ps->list[unifRand(0, ctx.ps->num)];
	    now  = get_now();
	    tmo  = now + RPC_TIMEOUT*1000000; 
	    xid  = ctx.req.hdr.xid;

	    out_test->poor = poor;
	    out_test->begin = now;
	    out_test->when = tmo;

	    // map xid to the state we need to keep
	    ctx.xidmap[xid] = out_test;
	
	    // insert a timeout event that keeps track of the RPC
	    { 
	      event new_ev(tmo, TEST_TMO);
	      new_ev.xid = xid;
	      ctx.eventq.insert(ev_pair(new_ev.wh, new_ev));
	    }
    
	    do_send(TEST_RPC_SIZE, &poor->addr, poor->name);
	    ++cntrs.test_req;
#ifdef CHATTY
	    dump_req(&ev, &ctx.req, poor);
#endif
	    break;

	case DUMP:

	    dump_counters();
	    // schedule the next one
	    schedule_dump();
	    break;

	case TEST_TMO:

	    if (!ctx.first_tmo) {
		printf("first timeout happened\n");
		ctx.first_tmo = true;
	    }
	    it = ctx.xidmap.find(ev.xid);
	    assert(it != ctx.xidmap.end());

	    out_test = it->second;
	    assert(out_test);
	    ctx.xidmap.erase(it); 

	    if (out_test->is_spam) {
		uint32 idx = out_test->n_spam - ctx.spam_beg;
		stamp_cntr* stmp_cntr = &ctx.stamp_cntrs[idx];
		stmp_cntr->tmo++;
	    }
	    free(out_test);

	    ++cntrs.test_tmo;
	    break;

	case PRELOAD_DONE:
	  
	    printf("preloading phase is over.\n");
	    msg = ";==PLD==\n";
	    log_to_latency(msg, strlen(msg), false);
	    ctx.preloading_done = true;
	    if (ctx.rfile == NULL) {
	      ctx.mean_interarr_us = 1000000.0 / (ctx.params->rate);	      
	      break;
	    } // else: fall through!!

        case RATE_UPDATE:
	    rate_update();
	    {
	      uint64 when = ev.wh + ctx.rtime * 1000000ll;
	      event new_ev(when, RATE_UPDATE);
	      printf("toi, in %llu, at %llu, now is %llu, last was %llu\n",
		     ctx.rtime, when, get_now(), ev.wh);
	      ctx.eventq.insert(ev_pair(new_ev.wh, new_ev));
	    }
	    break;

        case DONE_RUNNING:
	    print_and_exit();
	    break;

	default:
	    assert(false);
    }
}

/* 
 * The public interface to the send loop. Callers have to give a list of
 * portals,  a probability that a given stamp is reused, the size of the
 * universe of counters that can be spam and a rate to send TEST requests.
 */
void sendloop(config_params* params)
{
    fd_set read_fds;
    struct timeval tmo;
    char resp_buf[1500];
    bool zero_tmo;
    int rc;

    /* init stuff */
    FD_ZERO(&read_fds);
    setup_context(params);
    memset(&cntrs, 0, sizeof(cntrs));

    /* Need to schedule a test in the beginning to start the thing off. */
    schedule_test(get_now());
    schedule_dump();

    for (;;) {
   
	uint64 now   = get_now();
	// get the earliest event
	event ev     = ctx.eventq.begin()->second;
	int delta    = ev.wh - now;

	if (delta <= 0) {
	    zero_tmo = true;
	    tmo.tv_sec = tmo.tv_usec = 0;
	} else {
	    zero_tmo = false;
	    us2tv((uint32)delta, &tmo);
	}

	FD_ZERO(&read_fds);
	FD_SET(ctx.s, &read_fds);
	rc = select(ctx.s + 1, &read_fds, NULL, NULL, &tmo);

	if (rc == -1 && errno != EINTR) 
	    PERROR("select()");

	if (rc > 0 && FD_ISSET(ctx.s, &read_fds)) {
	    int len;
	    if ((len = recvfrom(ctx.s, resp_buf, 
				sizeof(resp_buf), 0, NULL, NULL)) < 0)
		PERROR("recvfrom()");
	    process_reply(resp_buf, len);
	}

	// if we're behind (if ev.wh < now) or if the select call timed
	// out, it's time to do the event.
	if (zero_tmo || rc == 0) {

	    ev_it it = ctx.eventq.begin(); 
	    ev = it->second;
	    ctx.eventq.erase(it);

	    process_event(ev);
	}

    }
}

#if 0
extern __inline__ unsigned long long int rdtsc()
{
    unsigned long long int x;
    __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
    return x;
}
#endif
