/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
/* 
* $Id: u2.c,v 1.21 2006/05/19 13:17:03 mwalfish Exp $
* 
*/
#include "u.h"
#include "send.h"

/* 
 * Reads in a list of portals in format hostname:port and returns
 * a pointer to a portals struct.
 */
portals* read_portals_list(char* portalfile)
{
    char buf[256];
    portals* all;
    FILE* f;
    int i;

    if ((f = fopen(portalfile, "r")) == NULL) 
	WORDY_ERROR("error opening file %s", portalfile);

    all = (portals*)malloc(sizeof(*all));
    assert(all);

    for (i = 0; fgets(buf, sizeof(buf), f) != NULL; i++) {

	/* nasty C 'parsing' */
	buf[strlen(buf)-1] = '\0';     /* remove newline */
	char* udp_port = strchr(buf, ':');
	*udp_port++ = '\0';

	char *udp_port_end = strchr(udp_port, ':');
	if (udp_port_end != NULL)
	  *udp_port_end = '\0';

	portal *p = (portal*)malloc(sizeof(*p));
	assert(p);

	memset(p, 0, sizeof(*p));
	strcpy(p->name, buf);
	p->port = atoi(udp_port);
	
	all->list[i] = p;
    }
    all->num = i;

    fclose(f);
    return all;
}

void resolve_hostnames(portals* ps)
{
    int i;
    portal* p;

    printf("resolving hostnames of the portals....\n");
    for (i = 0; i < ps->num; i++) {

	p = ps->list[i];

	/* set up sockaddr now; we'll keep using this */
	p->addr.sin_family = AF_INET;
	p->addr.sin_port  = htons(p->port);
        struct hostent* he = gethostbyname(p->name);	

	if (he == NULL) 
	    WORDY_ERROR("DNS lookup failed for host %s", p->name)
    	
	memcpy((char*)&p->addr.sin_addr.s_addr, 
	       he->h_addr_list[0], 
	       sizeof(long));
    }

}

portals* init_portals(char* portalfile)
{
    portals* ps;

    ps = read_portals_list(portalfile);
    resolve_hostnames(ps);

    return ps;
}


void dump_portals(portals* ps)
{
    for (int i = 0; i < ps->num; i++) {
	portal* p = ps->list[i];
	printf("%s:%d  (%s)\n", 
	          inet_ntoa(p->addr.sin_addr), 
		  p->port, 
		  p->name);
    }
}

void usage(char** argv)
{
    printf("usage: %s [-r num_times_to_repeat_each_counter] "
           "[-b begin_counter] "
           "[-e end_counter] "
           "[-o <t|s>] "
	   "[-l latency_out] "
	   "[-p preload_rate] "
	   "[-d] "
	   "[-f rate_file] "
	   "[-s rate_file_offset] "
	   "[-t rate_file_interval] "
	   "portal-list p_spam spam_universe TESTs/sec\n", argv[0]);
    printf(" -o t: do tests only\n"
           " -o s: do sets only\n"
	   "   -d: delay for rpc timeout after exhausting reuses\n");
    exit(1);
}

int main(int argc, char** argv)
{
    // required positional arguments
    int off_portal_list = 0;
    int off_spam        = 1;
    int off_sp_universe = 2;
    int off_tests       = 3;

    unsigned long msgc_begin = 0;               // message counter
    unsigned long msgc_max = 0xffffffff;  // maximum value of above counter
    bool tests_only = false;
    bool sets_only  = false;
    bool done_delay = false;
    int repeat = 1;
    char* latency_file = NULL;
    int preload_tests = -1;
    char *rate_file = NULL;
    unsigned long rate_file_o = 0;
    unsigned long long rate_file_t = 600LL;

    int opt;
    opterr = 0;
    while ( (opt = getopt(argc, argv, "b:e:o:r:l:p:f:s:t:d")) != -1) {

	switch (opt) {

	    case 'b':
		msgc_begin = atoi(optarg);
		break;

	    case 'e':
		msgc_max = atoi(optarg);
		break;

	    case 'o':
		switch (optarg[0]) {
		    case 't':
			tests_only = true;
			break;

		    case 's':
			sets_only = true;
			break;

		    default:
			printf("argument to 'o' must be "
			       "'t[est]' or 's[et]'\n");
			exit(1);
		}
		break;
	   
	    case 'r':
		repeat = atoi(optarg);
		break;

	    case 'l':
		latency_file = optarg;
		break;

	    case 'p':
	        preload_tests = atoi(optarg);
		break;

	    case 'f':
	        rate_file = optarg;
		break;

	    case 's':
	        rate_file_o = strtoul(optarg, NULL, 0);
		break;

	    case 't':
	        rate_file_t = strtoull(optarg, NULL, 0);
		break;

	    case 'd':
	        done_delay = true;
                break;

	    default:

		usage(argv);
	}
    }

    if (optind != argc - 4)
	usage(argv);

    int ind_portal_list = optind + off_portal_list;
    int ind_p_spam      = optind + off_spam;
    int ind_sp_universe = optind + off_sp_universe;
    int ind_tests       = optind + off_tests;

    char* portalfile = argv[ind_portal_list];

    char* dummy;
    double p_spam = strtod(argv[ind_p_spam], &dummy);
    if (dummy == argv[ind_p_spam]) usage(argv);

    uint32 sp_universe = strtol(argv[ind_sp_universe], &dummy, 10);
    if (dummy == argv[ind_sp_universe]) usage(argv);

    int tests_per_sec = strtol(argv[ind_tests], &dummy, 10);
    if (dummy == argv[ind_tests]) usage(argv);

    if (preload_tests == -1) preload_tests = tests_per_sec;

    portals* ps = init_portals(portalfile);
    
    printf("We will work with these portals:\n");
    dump_portals(ps);

    printf("parameters: TESTs/sec: %d. p_{spam}: %.3f. spam universe: %lu\n",
	   tests_per_sec, p_spam, sp_universe);

    /* some other initialization */
    signal(SIGINT, handle_sigint); 
    initRand(time(NULL));

    /* setup the configuration parameters, */
    config_params params;
    params.ps      = ps;
    params.pr_spam = p_spam;
    params.universe_sz = sp_universe;
    params.msgc_begin  = msgc_begin;
    params.msgc_max    = msgc_max;
    params.repeat      = repeat;
    params.rate        = tests_per_sec;
    params.preload_rate= preload_tests;
    params.done_delay  = done_delay;
    params.tests_only  = tests_only;
    params.sets_only   = sets_only;
    params.latency_file = latency_file;
    params.rate_file   = rate_file;
    params.rate_file_o = rate_file_o;
    params.rate_file_t = rate_file_t;

    /* and go . . . */
    sendloop(&params);

    exit(0);
}

extern void print_and_exit();

void handle_sigint(int signal)
{
    printf("caught SIGINT, exiting.....\n");
    stop_running();
}
